# -*- coding: utf-8 -*-

############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
# Modifications copyright (c) 2019-2020 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import os
from keyword import kwlist
from collections import namedtuple, Counter
from itertools import compress, islice
from functools import partial
from contextlib import contextmanager
from operator import itemgetter
from math import isnan

from accelerator.compat import unicode, uni, ifilter, imap, iteritems
from accelerator.compat import builtins, open, getarglist, izip, izip_longest
from accelerator.compat import str_types, int_types, FileNotFoundError

from accelerator import blob
from accelerator.extras import DotDict, job_params, _ListTypePreserver
from accelerator.job import Job
from accelerator.dsutil import typed_writer, _type2iter
from accelerator.error import NoSuchDatasetError, DatasetUsageError, DatasetError

kwlist = set(kwlist)
# Add some keywords that are not in all versions
kwlist.update({
	'exec', 'print',                                       # only in py2
	'False', 'None', 'True', 'nonlocal', 'async', 'await', # only in py3
})

iskeyword = frozenset(kwlist).__contains__

# A dataset is defined by a pickled DotDict containing at least the following (all strings are unicode):
#     version = (3, 1,),
#     filename = "filename" or None,
#     hashlabel = "column name" or None,
#     caption = "caption",
#     columns = {"column name": DatasetColumn,},
#     previous = "previous_jid/datasetname" or None,
#     parent = "parent_jid/datasetname" or None,
#     lines = [line, count, per, slice,],
#     cache = ((id, data), ...), # key is missing if there is no cache in this dataset
#     cache_distance = datasets_since_last_cache, # key is missing if previous is None
#
# A DatasetColumn has these fields:
#     type = "type", # something that exists in type2iter and doesn't start with _
#     backing_type = "type", # something that exists in type2iter (v2 used type for this)
#                              (as the support for v2 has been removed, backing_type will
#                              always be the same as type. until something changes again.)
#     location = something, # where the data for this column lives
#         in version 2 and 3 this is "jobid/path/to/file" if .offsets else "jobid/path/with/%s/for/sliceno"
#     min = minimum value in this dataset or None
#     max = maximum value in this dataset or None
#     offsets = (offset, per, slice) or None for non-merged slices.
#     none_support = bool # not present in version 3.0, implicitly True there except for bits-types.
#
# Going from a DatasetColumn to a filename is like this for version 2 and 3 datasets:
#     jid, path = dc.location.split('/', 1)
#     jid = Job(jid)
#     if dc.offsets:
#         jid.filename(path)
#         seek to dc.offsets[sliceno], read only ds.lines[sliceno] values.
#     else:
#         jid.filename(path % sliceno)
# There is a ds.column_filename function to do this for you (not the seeking, obviously).
#
# The dataset pickle is jid/name/dataset.pickle, so jid/default/dataset.pickle for the default dataset.

def _clean_name(n, seen_n):
	n = ''.join(c if c.isalnum() else '_' for c in n)
	if not n or n[0].isdigit():
		n = '_' + n
	while n in seen_n or iskeyword(n):
		n += '_'
	seen_n.add(n)
	return n

def _dsid(t):
	if not t:
		return None
	if isinstance(t, (tuple, list)):
		jid, name = t
		if not jid:
			return None
		t = '%s/%s' % (jid.split('/')[0], uni(name) or 'default')
	elif isinstance(t, DatasetWriter):
		from accelerator.g import job
		t = '%s/%s' % (job, t.name)
	if '/' not in t:
		t += '/default'
	return uni(t)

# If we want to add fields to later versions, using a versioned name will
# allow still loading the old versions without messing with the constructor.
_DatasetColumn_3_2 = namedtuple('_DatasetColumn_3_2', 'type backing_type location min max offsets none_support')
DatasetColumn = _DatasetColumn_3_2
# It's probably usually best to generate the new type so the rest of the code needs no special handling.
class _DatasetColumn_3_1(object):
	def __new__(cls, type, backing_type, name, location, min, max, offsets, none_support):
		return _DatasetColumn_3_2(type, backing_type, location, min, max, offsets, none_support)
class _DatasetColumn_3_0(object):
	def __new__(cls, type, backing_type, name, location, min, max, offsets):
		none_support = not backing_type.startswith('bits')
		return _DatasetColumn_3_1(type, backing_type, name, location, min, max, offsets, none_support)

class _New_dataset_marker(unicode): pass
_new_dataset_marker = _New_dataset_marker('new')
_no_override = object()

_ds_cache = {}
def _ds_load(obj):
	n = unicode(obj)
	if n not in _ds_cache:
		fn = obj.job.filename(obj._name('pickle'))
		if not os.path.exists(fn):
			raise NoSuchDatasetError('Dataset %r does not exist' % (n,))
		_ds_cache[n] = blob.load(fn)
		_ds_cache.update(_ds_cache[n].get('cache', ()))
	return _ds_cache[n]

def _namechk(name):
	if not name:
		raise DatasetUsageError("Dataset name can't be empty")
	name = uni(name)
	for c in '/\n\0%':
		if c in name:
			raise DatasetUsageError("Dataset name can't contain " + repr(c))
	return name

_dummy_iter = iter(())

class Dataset(unicode):
	"""
	Represents a dataset. Is also a string 'jobid/name', or just 'jobid' if
	name is 'default' (for better backwards compatibility).

	You usually don't have to make these yourself, because datasets.foo is
	already a Dataset instance (or None).

	You can pass jobid="jid/name" or jobid="jid", name="name", or skip
	name completely for "default".

	You can also pass jobid={jid: dsname} to resolve dsname from the datasets
	passed to jid. This gives None if that option was unset.

	These decay to a (unicode) string when pickled.
	"""

	def __new__(cls, jobid, name=None):
		if isinstance(jobid, (tuple, list)):
			jobid = _dsid(jobid)
		elif isinstance(jobid, dict):
			if name:
				raise DatasetUsageError("Don't pass both a separate name and jobid as {job: dataset}")
			if len(jobid) != 1:
				raise DatasetUsageError("Only pass a single {job: dataset}")
			jobid, dsname = next(iteritems(jobid))
			if not jobid:
				return None
			jobid = job_params(jobid, default_empty=True).datasets.get(dsname)
			if not jobid:
				return None
		if not jobid:
			raise DatasetUsageError("If you really meant to use yourself as a dataset, pass your jobid explicitly.")
		if '/' in jobid:
			if name:
				raise DatasetUsageError("Don't pass both a separate name and jobid as jid/name")
			jobid, name = jobid.split('/', 1)
		name = _namechk(name or 'default')
		if name == 'default':
			suffix = ''
		else:
			suffix = '/' + name
		if jobid is _new_dataset_marker:
			from accelerator.g import job
			fullname = job + suffix
		else:
			fullname = jobid + suffix
		obj = unicode.__new__(cls, fullname)
		obj.name = uni(name or 'default')
		if jobid is _new_dataset_marker:
			obj._data = DotDict({
				'version': (3, 0,),
				'filename': None,
				'hashlabel': None,
				'caption': '',
				'columns': {},
				'parent': None,
				'previous': None,
				'lines': [],
			})
			obj.job = None
		else:
			obj.job = Job(jobid)
			obj._data = DotDict(_ds_load(obj))
			if obj._data.version[0] != 3:
				raise DatasetError("%s/%s: Unsupported dataset pickle version %r" % (jobid, name, obj._data.version,))
			obj._data.columns = dict(obj._data.columns)
		obj._cache = {}
		return obj

	# Look like a string after pickling
	def __reduce__(self):
		return unicode, (unicode(self),)

	@property
	def columns(self):
		"""{name: DatasetColumn}"""
		return self._data.columns

	def _cached(self, thing):
		org_value = self._data[thing]
		if not org_value:
			return
		if self._cache.get(thing, (None,))[0] != org_value:
			if isinstance(org_value, tuple):
				value = tuple(Dataset(v) for v in org_value)
			else:
				value = Dataset(org_value)
			self._cache[thing] = (org_value, value)
		return self._cache[thing][1]

	@property
	def previous(self):
		return self._cached('previous')

	@property
	def parent(self):
		return self._cached('parent')

	@property
	def filename(self):
		return self._data.filename

	@property
	def hashlabel(self):
		return self._data.hashlabel

	@property
	def caption(self):
		return self._data.caption

	@property
	def lines(self):
		return self._data.lines

	@property
	def shape(self):
		return (len(self.columns), sum(self.lines),)

	def link_to_here(self, name='default', column_filter=None, override_previous=_no_override, filename=None):
		"""Use this to expose a subjob as a dataset in your job:
		Dataset(subjid).link_to_here()
		will allow access to the subjob dataset under your jid.
		Specify column_filter as an iterable of columns to include
		if you don't want all of them.
		Use override_previous to rechain (or unchain) the dataset.
		You can change the filename too, or clear it by setting ''.
		"""
		d = Dataset(self)
		if column_filter:
			column_filter = set(column_filter)
			filtered_columns = {k: v for k, v in d._data.columns.items() if k in column_filter}
			left_over = column_filter - set(filtered_columns)
			if left_over:
				raise DatasetUsageError("Columns in filter not available in dataset: %r" % (left_over,))
			if not filtered_columns:
				raise DatasetUsageError("Filter produced no desired columns.")
			d._data.columns = filtered_columns
		from accelerator.g import job
		if override_previous is not _no_override:
			override_previous = _dsid(override_previous)
			if override_previous:
				# make sure it's valid
				Dataset(override_previous)
			d._data.previous = override_previous
			d._update_caches()
		d._data.parent = '%s/%s' % (d.job, d.name,)
		if filename is not None:
			d._data.filename = filename or None
		d.job = job
		d.name = uni(name)
		d._save()
		_datasets_written.append(d.name)
		return Dataset(d.job, d.name)

	def merge(self, other, name='default', previous=None, allow_unrelated=False):
		"""Merge this and other dataset. Columns from other take priority.
		If datasets do not have a common ancestor you get an error unless
		allow_unrelated is set. The new dataset always has the previous
		specified here (even if None).
		Returns the new dataset.
		"""
		from accelerator.g import job
		new_ds = Dataset(self)
		other = Dataset(other)
		name = _namechk(name)
		if previous:
			previous = Dataset(previous)
		else:
			previous = None
		if self == other:
			raise DatasetUsageError("Can't merge with myself (%s)" % (other,))
		if self.lines != other.lines:
			raise DatasetUsageError("%s and %s don't have the same line counts" % (self, other,))
		hashlabels = {self.hashlabel, other.hashlabel}
		hashlabels.discard(None)
		if len(hashlabels) > 1:
			raise DatasetUsageError("Hashlabel mismatch, %s has %r, %s has %r" % (self, self.hashlabel, other, other.hashlabel,))
		new_ds._data.columns.update(other._data.columns)
		if not allow_unrelated:
			def parents(ds, tips):
				if not isinstance(ds, tuple):
					ds = (ds,)
				for p in ds:
					pds = Dataset(p)
					if pds.parent:
						parents(pds.parent, tips)
					else:
						tips.add(pds)
				return tips
			related = parents(self, set()) & parents(other, set())
			if not related:
				raise DatasetUsageError("%s and %s have no common ancenstors, set allow_unrelated to allow this" % (self, other,))
		new_ds.job = job
		new_ds.name = name
		new_ds._data.previous = previous
		new_ds._data.parent = (self, other)
		new_ds._data.filename = None
		new_ds._data.caption = None
		new_ds._update_caches()
		new_ds._save()
		_datasets_written.append(name)
		return job.dataset(name) # new_ds has the wrong string value, so we must make a new instance here.

	def _column_iterator(self, sliceno, col, _type=None, **kw):
		if sliceno is not None and self.lines[sliceno] == 0:
			return _dummy_iter
		dc = self.columns[col]
		mkiter = partial(_type2iter[_type or dc.backing_type], **kw)
		def one_slice(sliceno):
			fn = self.column_filename(col, sliceno)
			if dc.offsets:
				return mkiter(fn, seek=dc.offsets[sliceno], want_count=self.lines[sliceno])
			else:
				return mkiter(fn, want_count=self.lines[sliceno])
		if sliceno is None:
			from accelerator.g import slices
			from itertools import chain
			return chain(*[one_slice(s) for s in range(slices)])
		else:
			return one_slice(sliceno)

	def _iterator(self, sliceno, columns=None, copy_mode=False):
		res = []
		not_found = []
		for col in columns or sorted(self.columns):
			if col in self.columns:
				if copy_mode:
					t = _copy_mode_overrides.get(self.columns[col].backing_type)
					res.append(self._column_iterator(sliceno, col, _type=t))
				else:
					res.append(self._column_iterator(sliceno, col))
			else:
				not_found.append(col)
		if not_found:
			raise DatasetError("Columns %r not found in %s/%s" % (not_found, self.job, self.name,))
		return res

	def _hashfilter(self, sliceno, hashlabel, it):
		from accelerator.g import slices
		return compress(it, self._column_iterator(None, hashlabel, hashfilter=(sliceno, slices)))

	def column_filename(self, colname, sliceno=None):
		dc = self.columns[colname]
		jid, name = dc.location.split('/', 1)
		jid = Job(jid)
		if dc.offsets:
			return jid.filename(name)
		else:
			if sliceno is None:
				sliceno = '%s'
			return jid.filename(name % (sliceno,))

	def chain(self, length=-1, reverse=False, stop_ds=None):
		if stop_ds:
			# resolve all formats to the same format
			stop_ds = Dataset(stop_ds)
		chain = DatasetChain()
		current = self
		while length != len(chain) and current != stop_ds:
			chain.append(current)
			if not current.previous:
				break
			current = Dataset(current.previous)
		if not reverse:
			chain.reverse()
		return chain

	def iterate_chain(self, sliceno, columns=None, length=-1, range=None, sloppy_range=False, reverse=False, hashlabel=None, stop_ds=None, pre_callback=None, post_callback=None, filters=None, translators=None, status_reporting=True, rehash=False, slice=None, copy_mode=False):
		"""Iterate a list of datasets. See .chain and .iterate_list for details."""
		chain = self.chain(length, reverse, stop_ds)
		return self.iterate_list(sliceno, columns, chain, range=range, sloppy_range=sloppy_range, hashlabel=hashlabel, pre_callback=pre_callback, post_callback=post_callback, filters=filters, translators=translators, status_reporting=status_reporting, rehash=rehash, slice=slice, copy_mode=copy_mode)

	def iterate(self, sliceno, columns=None, hashlabel=None, pre_callback=None, post_callback=None, filters=None, translators=None, status_reporting=True, rehash=False, slice=None, copy_mode=False):
		"""Iterate just this dataset. See .iterate_list for details."""
		return self.iterate_list(sliceno, columns, [self], hashlabel=hashlabel, pre_callback=pre_callback, post_callback=post_callback, filters=filters, translators=translators, status_reporting=status_reporting, rehash=rehash, slice=slice, copy_mode=copy_mode)

	@staticmethod
	def iterate_list(sliceno, columns, datasets, range=None, sloppy_range=False, hashlabel=None, pre_callback=None, post_callback=None, filters=None, translators=None, status_reporting=True, rehash=False, slice=None, copy_mode=False):
		"""Iterator over the specified columns from datasets
		(iterable of dataset-specifiers, or single dataset-specifier).
		callbacks are called before and after each dataset is iterated.

		filters decide which rows to include and can be a callable
		(called with the candidate tuple), or a dict {name: filter}.
		In the latter case each individual filter is called with the column
		value, or if it's None uses the column value directly.
		All filters must say yes to yield a row.
		examples:
		filters={'some_col': some_dict.get}
		filters={'some_col': some_set.__contains__}
		filters={'some_col': some_str.__eq__}
		filters=lambda line: line[0] == line[1]

		translators transform data values. It can be a callable (called with the
		candidate tuple and expected to return a tuple of the same length) or a
		dict {name: translation}.
		Each translation can be a function (called with the column value and
		returning the new value) or dict. Items missing in the dict yield None,
		which can be removed with filters={'col': None}.

		Translators run before filters.

		You can also pass a single name (a str) as columns, in which case you
		don't get a tuple back (just the values). Tuple-filters/translators also
		get just the value in this case (column versions are unaffected).

		If you pass a false value for columns you get all columns in name order.

		If you pass sliceno=None you get all slices.
		If you pass sliceno="roundrobin" you also get all slices, but one value
		at a time across slices. (This can be used to iterate a csvimport in the
		order of the original file.)

		If you specify a hashlabel and rehash=False (the default) you will
		get an error if the a dataset does not use the specified hashlabel.
		If you specify rehash=True such datasets will be rehashed during
		iteration. You should usually build a new rehashed dataset (using
		the dataset_hashpart method), but this is available for when it makes
		sense.

		range limits which rows you see. Specify {colname: (start, stop)} and
		only rows where start <= colvalue < stop will be returned.
		If you set sloppy_range=True you may get all rows from datasets that
		contain any rows you asked for. (This can be faster.)

		status_reporting should normally be left as True, which will give you
		information about this iteration in ^T, but there is one case where you
		need to turn it off:
		If you manually zip a bunch of iterators, only one should do status
		reporting. (Otherwise it looks like you have nested iteration in ^T,
		and you will get warnings about incorrect ending order of statuses.)

		slice takes a slice object defining which lines you want returned.
		Negative offsets are allowed. Both negative and positive offsets
		outside the range of the specified datasets give an error.
		This is equivalent to islice(iterate_list(...), start, stop, step)
		except it's faster, allows negative offsets and errors on too big
		offsets.
		For convenience you can also pass slice=<some int> which is equivalent
		to slice(<some int>, None). Note that this is not the same as
		slice(<some int>), but more useful here.

		copy_mode makes no promises about the returned types.
		Use it together with copy_mode on a DatasetWriter for faster copying.
		Not compatible with columns changing types across the list.
		Also not compatible with filters or translators.
		"""

		if isinstance(datasets, str_types + (Dataset, dict)):
			datasets = [datasets]
		datasets = [ds if isinstance(ds, Dataset) else Dataset(ds) for ds in datasets]
		slices = len(datasets[0].lines)
		if columns is None:
			columns = datasets[0].columns
		if isinstance(columns, str_types):
			columns = [columns]
			want_tuple = False
		else:
			if isinstance(columns, dict):
				columns = sorted(columns)
			want_tuple = True
		if copy_mode:
			assert not filters, "copy_mode is not compatible with filters"
			assert not translators, "copy_mode is not compatible with translators"
			need_types = {col: datasets[0].columns[col].type for col in columns}
			for d in datasets:
				for col, t in need_types.items():
					assert d.columns[col].type == t, "%s column %s has type %s, not %s" % (d, col, d.columns[col].type, t,)
		to_iter = []
		if range:
			if len(range) != 1:
				raise DatasetUsageError("Specify exactly one range column.")
			range_k, (range_bottom, range_top,) = next(iteritems(range))
			if range_bottom is None and range_top is None:
				range = None
		if slice:
			if isinstance(slice, int_types):
				slice = builtins.slice(slice, None, 1)
			slice = builtins.slice(slice.start or 0, slice.stop, slice.step or 1)
			if slice.step <= 0:
				raise DatasetUsageError("only positive slice steps are supported")
			if slice.start < 0 or (slice.stop or 0) < 0:
				msg = "slice with negative index is not supported with "
				if range: raise DatasetUsageError(msg + "range")
				if rehash: raise DatasetUsageError(msg + "rehash")
				if filters: raise DatasetUsageError(msg + "filters")
			if isinstance(sliceno, int_types):
				total_lines = sum(d.lines[sliceno] for d in datasets)
			else:
				total_lines = sum(sum(d.lines) for d in datasets)
			if slice.start < 0:
				if -slice.start > total_lines:
					raise DatasetUsageError("Wanted last %d lines, but only %d lines available" % (-slice.start, total_lines,))
				slice = builtins.slice(total_lines + slice.start, slice.stop, slice.step)
			if (slice.stop or 0) < 0:
				if -slice.stop > total_lines:
					raise DatasetUsageError("Wanted to stop %d lines before end, but only %d lines available" % (-slice.stop, total_lines,))
				slice = builtins.slice(slice.start, total_lines + slice.stop, slice.step)
			if slice.start > total_lines:
				raise DatasetUsageError("Wanted to skip %d lines, but only %d lines available" % (slice.start, total_lines,))
			if (slice.stop or 0) > total_lines:
				raise DatasetUsageError("Wanted to stop after %d lines, but only %d lines available" % (slice.stop, total_lines,))
			if slice.start == total_lines:
				return iter(())
			if slice.stop is not None:
				if slice.stop < slice.start:
					raise DatasetUsageError("Wanted %r, but start is bigger than stop" % (slice,))
				if slice.start == slice.stop:
					return iter(())
			def adj_slice(lines):
				stop = slice.stop
				if stop is not None:
					stop -= lines
				return builtins.slice(slice.start - lines, stop, slice.step)
			slice_skipping = True
			slice_first = True
			if slice.start == 0 and slice.stop is None and slice.step == 1:
				slice = None
		if not slice:
			slice_skipping = False
			slice_first = False
		for d in datasets:
			if isinstance(sliceno, int_types):
				lines_here = d.lines[sliceno]
			else:
				lines_here = sum(d.lines)
			if lines_here == 0:
				continue
			if slice_skipping and slice.start >= lines_here:
				slice = adj_slice(lines_here)
				continue
			else:
				slice_skipping = False
			if range:
				c = d.columns[range_k]
				if range_top is not None and c.min >= range_top:
					continue
				if range_bottom is not None and c.max < range_bottom:
					continue
			if hashlabel is not None and d.hashlabel != hashlabel:
				if not rehash:
					raise DatasetUsageError("%s has hashlabel %r, not %r" % (d, d.hashlabel, hashlabel,))
				if hashlabel not in d.columns:
					raise DatasetUsageError("Can't rehash %s on non-existant column %r" % (d, hashlabel,))
				rehash_on = hashlabel
			else:
				rehash_on = None
			if sliceno is None:
				if slice_first:
					for first_slice in builtins.range(slices):
						if slice.start < d.lines[first_slice]:
							break
						slice = adj_slice(d.lines[first_slice])
				else:
					first_slice = 0
				for ix in builtins.range(first_slice, slices):
					# Ignore rehashing - order is generally not guaranteed with sliceno=None
					to_iter.append((d, ix, None,))
			else:
				to_iter.append((d, sliceno, rehash_on,))
			slice_first = False
		if not rehash and not pre_callback and not post_callback and not isinstance(sliceno, str_types):
			# No reason to include empty slices then
			to_iter = [t for t in to_iter if t[0].lines[t[1]]]
		filter_func = Dataset._resolve_filters(columns, filters, want_tuple)
		translation_func, translators = Dataset._resolve_translators(columns, translators)
		if sloppy_range:
			range = None
		from itertools import chain
		kw = dict(
			columns=columns,
			pre_callback=pre_callback,
			post_callback=post_callback,
			filter_func=filter_func,
			translation_func=translation_func,
			translators=translators,
			want_tuple=want_tuple,
			range=range,
			status_reporting=status_reporting,
			copy_mode=copy_mode,
		)
		if sliceno == "roundrobin":
			# We do our own status reporting
			kw["status_reporting"] = False
			def rr_inner(d, rehash):
				todo = []
				for ix in builtins.range(slices):
					part = (d, ix, rehash)
					todo.append(chain.from_iterable(Dataset._iterate_datasets([part], **kw)))
				fv = object() # unique marker
				return (
					v for v in
					chain.from_iterable(izip_longest(*todo, fillvalue=fv))
					if v is not fv
				)
			def rr_outer():
				with Dataset._iterstatus(status_reporting, to_iter) as update:
					for ix, (d, sliceno, rehash) in enumerate(to_iter, 1):
						update(ix, d, sliceno, rehash)
						yield rr_inner(d, rehash)
			res = chain.from_iterable(rr_outer())
		else:
			res = chain.from_iterable(Dataset._iterate_datasets(to_iter, **kw))
		if slice and (slice.start or slice.stop or slice.step > 1):
			res = islice(res, slice.start, slice.stop, slice.step)
		return res

	@staticmethod
	def _resolve_filters(columns, filters, want_tuple):
		if filters and not callable(filters):
			# Sort in column order, to allow selecting an efficient order.
			filters = sorted((columns.index(name), f,) for name, f in filters.items())
			if not want_tuple:
				return filters[0][1] or bool
			# Build "lambda t: f0(t[0]) and f1(t[1]) and ..."
			fs = []
			arg_n = []
			arg_v = []
			for ix, f in filters:
				if f is None or f is bool:
					# use value directly
					fs.append('t[%d]' % (ix,))
				else:
					n = 'f%d' % (ix,)
					arg_n.append(n)
					arg_v.append(f)
					fs.append('%s(t[%d])' % (n, ix,))
			f = 'lambda t: ' + ' and '.join(fs)
			# Add another lambda to put all fN into local variables.
			# (This is faster than putting them in "locals", you get
			# LOAD_DEREF instead of LOAD_GLOBAL.)
			f = 'lambda %s: %s' % (', '.join(arg_n), f)
			return eval(f, {}, {})(*arg_v)
		else:
			return filters

	@staticmethod
	def _resolve_translators(columns, translators):
		if not translators:
			return None, {}
		if callable(translators):
			return translators, {}
		else:
			res = {}
			for name, f in translators.items():
				if not callable(f):
					f = f.get
				res[columns.index(name)] = f
			return None, res

	@staticmethod
	@contextmanager
	def _iterstatus(status_reporting, to_iter):
		if not status_reporting:
			yield lambda *_: None
			return
		from accelerator.statmsg import status
		def fmt_dsname(d, sliceno, rehash):
			if rehash is not None:
				return d + ':REHASH'
			else:
				return '%s:%s' % (d, sliceno)
		if len(to_iter) == 1:
			msg_head = 'Iterating ' + fmt_dsname(*to_iter[0])
			def update_status(ix, d, sliceno, rehash):
				pass
		else:
			msg_head = 'Iterating %s to %s' % (fmt_dsname(*to_iter[0]), fmt_dsname(*to_iter[-1]),)
			def update_status(ix, d, sliceno, rehash):
				update('%s, %d/%d (%s)' % (msg_head, ix, len(to_iter), fmt_dsname(d, sliceno, rehash)))
		with status(msg_head) as update:
			yield update_status

	@staticmethod
	def _iterate_datasets(to_iter, columns, pre_callback, post_callback, filter_func, translation_func, translators, want_tuple, range, status_reporting, copy_mode):
		skip_ds = None
		def argfixup(func, is_post):
			if func:
				if len(getarglist(func)) == 1:
					seen_ds = [None]
					def wrapper(d, sliceno=None):
						if d != seen_ds[0]:
							if is_post:
								if seen_ds[0] and seen_ds[0] != skip_ds:
									func(seen_ds[0])
							else:
								func(d)
							seen_ds[0] = d
					return wrapper, True
			return func, False
		pre_callback, unsliced_pre_callback = argfixup(pre_callback, False)
		post_callback, unsliced_post_callback = argfixup(post_callback, True)
		if not to_iter:
			return
		if range:
			range_k, (range_bottom, range_top,) = next(iteritems(range))
			range_check = range_check_function(range_bottom, range_top)
			if range_k in columns and range_k not in translators and not translation_func:
				has_range_column = True
				range_i = columns.index(range_k)
				if want_tuple:
					range_f = lambda t: range_check(t[range_i])
				else:
					range_f = range_check
			else:
				has_range_column = False
		with Dataset._iterstatus(status_reporting, to_iter) as update:
			for ix, (d, sliceno, rehash) in enumerate(to_iter, 1):
				if unsliced_post_callback:
					try:
						post_callback(d)
					except StopIteration:
						return
				update(ix, d, sliceno, rehash)
				if pre_callback:
					if d == skip_ds:
						continue
					try:
						pre_callback(d, sliceno)
					except SkipSlice:
						if unsliced_pre_callback:
							skip_ds = d
						continue
					except SkipDataset:
						skip_ds = d
						continue
					except StopIteration:
						return
				it = d._iterator(None if rehash is not None else sliceno, columns, copy_mode=copy_mode)
				for ix, trans in translators.items():
					it[ix] = imap(trans, it[ix])
				if want_tuple:
					it = izip(*it)
				else:
					it = it[0]
				if rehash is not None:
					it = d._hashfilter(sliceno, rehash, it)
				if translation_func:
					it = imap(translation_func, it)
				if range:
					c = d.columns[range_k]
					if c.min is not None and (not range_check(c.min) or not range_check(c.max)):
						if has_range_column:
							it = ifilter(range_f, it)
						else:
							if rehash is not None:
								filter_it = d._hashfilter(sliceno, rehash, d._column_iterator(None, range_k))
							else:
								filter_it = d._column_iterator(sliceno, range_k)
							it = compress(it, imap(range_check, filter_it))
				if filter_func:
					it = ifilter(filter_func, it)
				yield it
				if post_callback and not unsliced_post_callback:
					try:
						post_callback(d, sliceno)
					except StopIteration:
						return
			if unsliced_post_callback:
				try:
					post_callback(None)
				except StopIteration:
					return

	@staticmethod
	def new(columns, filenames, lines, minmax={}, filename=None, hashlabel=None, caption=None, previous=None, name='default'):
		"""columns = {"colname": "type"}, lines = [n, ...] or {sliceno: n}"""
		columns = {uni(k): (uni(v[0]), bool(v[1])) if isinstance(v, tuple) else (uni(v), False) for k, v in columns.items()}
		if hashlabel is not None:
			hashlabel = uni(hashlabel)
			if hashlabel not in columns:
				raise DatasetUsageError("Hashlabel (%r) does not exist" % (hashlabel,))
		res = Dataset(_new_dataset_marker, name)
		res._data.lines = list(Dataset._linefixup(lines))
		res._data.hashlabel = hashlabel
		res._append(columns, filenames, minmax, filename, caption, previous, None, name)
		return res

	@staticmethod
	def _linefixup(lines):
		from accelerator.g import slices
		if isinstance(lines, dict):
			if set(lines) != set(range(slices)):
				raise DatasetUsageError("Lines must be specified for all slices")
			lines = [c for _, c in sorted(lines.items())]
		if len(lines) != slices:
			raise DatasetUsageError("Lines must be specified for all slices")
		return lines

	def append(self, columns, filenames, lines, minmax={}, filename=None, hashlabel=None, hashlabel_override=False, caption=None, previous=None, column_filter=None, name='default'):
		hashlabel = uni(hashlabel)
		if hashlabel_override:
			self._data.hashlabel = hashlabel
		elif hashlabel is not None and self.hashlabel != hashlabel:
			raise DatasetUsageError("Hashlabel mismatch %r != %r" % (self.hashlabel, hashlabel,))
		if self._linefixup(lines) != self.lines:
			raise DatasetUsageError("New columns don't have the same number of lines as parent columns")
		columns = {uni(k): (uni(v[0]), bool(v[1])) if isinstance(v, tuple) else (uni(v), False) for k, v in columns.items()}
		self._append(columns, filenames, minmax, filename, caption, previous, column_filter, name)

	def _minmax_merge(self, minmax):
		def minmax_fixup(a, b):
			res_min = a[0]
			if res_min is None: res_min = b[0]
			res_max = a[1]
			if res_max is None: res_max = b[1]
			return [res_min, res_max]
		def nanfix(cmp, a, b):
			if isinstance(a, float) and isnan(a):
				return b
			elif isinstance(b, float) and isnan(b):
				return a
			else:
				return cmp(a, b)
		res = {}
		for part in minmax.values():
			for name, mm in part.items():
				mm = tuple(mm)
				if mm != (None, None):
					omm = minmax_fixup(res.get(name, (None, None,)), mm)
					mm = minmax_fixup(mm, omm)
					res[name] = [nanfix(min, mm[0], omm[0]), nanfix(max, mm[1], omm[1])]
		return res

	def _append(self, columns, filenames, minmax, filename, caption, previous, column_filter, name):
		from accelerator.g import job
		name = uni(name)
		filenames = {uni(k): uni(v) for k, v in filenames.items()}
		if set(columns) != set(filenames):
			raise DatasetUsageError("columns and filenames don't have the same keys")
		if self.job and (self.job != job or self.name != name):
			self._data.parent = '%s/%s' % (self.job, self.name,)
		self.job = job
		self.name = name
		self._data.filename = uni(filename) or self._data.filename or None
		self._data.caption  = uni(caption) or self._data.caption or uni(job)
		self._data.previous = _dsid(previous)
		for n in ('cache', 'cache_distance'):
			if n in self._data: del self._data[n]
		minmax = self._minmax_merge(minmax)
		if column_filter:
			column_filter = set(column_filter)
			filtered_columns = {k: v for k, v in self._data.columns.items() if k in column_filter}
			left_over = column_filter - set(filtered_columns)
			if left_over:
				raise DatasetUsageError("Columns in filter not available in dataset: %r" % (left_over,))
			self._data.columns = filtered_columns
		for n, (t, none_support) in sorted(columns.items()):
			if t not in _type2iter:
				raise DatasetUsageError('Unknown type %s on column %s' % (t, n,))
			mm = minmax.get(n, (None, None,))
			t = uni(t)
			self._data.columns[n] = DatasetColumn(
				type=t,
				backing_type=t,
				location='%s/%s/%%s.%s' % (job, self.name, filenames[n]),
				min=mm[0],
				max=mm[1],
				offsets=None,
				none_support=none_support,
			)
			self._maybe_merge(n)
		self._update_caches()
		self._save()

	def _update_caches(self):
		for k in ('cache', 'cache_distance',):
			if k in self._data:
				del self._data[k]
		try:
			d = self.previous
		except NoSuchDatasetError:
			j, n = self._data['previous'].split('/', 1)
			if self.job == j and n in _datasetwriters:
				raise DatasetUsageError('previous dataset %r must be .finish()ed first.' % (n,))
			else:
				raise
		if d:
			cache_distance = d._data.get('cache_distance', 1) + 1
			if cache_distance == 64:
				cache_distance = 0
				chain = self.chain(64)
				self._data['cache'] = tuple((unicode(d), d._data) for d in chain[:-1])
			self._data['cache_distance'] = cache_distance

	def _maybe_merge(self, n):
		from accelerator.g import slices
		if slices < 2:
			return
		fn = self.column_filename(n)
		def getsize(sliceno):
			try:
				return os.path.getsize(fn % (sliceno,))
			except FileNotFoundError:
				return None
		sizes = [(sliceno, getsize(sliceno)) for sliceno, cnt in enumerate(self.lines)]
		bare_sizes = [z for _, z in sizes if z]
		if sum(bare_sizes) / (len(bare_sizes) or 1) > 524288: # arbitrary guess of good size
			return
		offsets = []
		pos = 0
		with open(fn % ('m',), 'wb') as m_fh:
			for sliceno, size in sizes:
				if size:
					with open(fn % (sliceno,), 'rb') as p_fh:
						data = p_fh.read()
					assert len(data) == size, "Slice %d is %d bytes, not %d?" % (sliceno, len(data), size,)
					m_fh.write(data)
				offsets.append(pos)
				if size is not None:
					os.unlink(fn % (sliceno,))
					pos += size
		c = self._data.columns[n]
		self._data.columns[n] = c._replace(
			offsets=offsets,
			location=c.location % ('m',),
		)

	def _save(self):
		if not os.path.exists(self.name):
			os.mkdir(self.name)
		blob.save(self._data, self._name('pickle'), temp=False, _hidden=True)

	def _name(self, thing):
		return '%s/dataset.%s' % (self.name, thing,)

_datasetwriters = {}
_datasets_written = []

_nodefault = object()

_copy_mode_overrides = dict.fromkeys(('unicode', 'ascii', 'json', 'pickle'), 'bytes')

# short non-colliding filenames safe for any filesystem
def _fngen():
	from itertools import cycle
	chars = cycle('abcdefghijklmnopqrstuvwxyz0123456789_')
	prefixgen = _fngen()
	prefix = ""
	while True:
		c = next(chars)
		yield prefix + c
		if c == '_':
			prefix = next(prefixgen)

class DatasetWriter(object):
	"""
	Create in prepare, use in analysis. Or do the whole thing in
	synthesis.

	You can pass these through prepare_res, or get them by trying to
	create a new writer in analysis (don't specify any arguments except
	an optional name).

	There are three writing functions with different arguments:

	dw.write_dict({column: value})
	dw.write_list([value, value, ...])
	dw.write(value, value, ...)

	Values are in the same order as you add()ed the columns (which is in
	sorted order if you passed a dict). The dw.write() function names the
	arguments from the columns too.

	If you set a column type to None that column is not inherited from the
	parent dataset. (Only works as an init argument, not with dw.add.)

	If you want support for None values in a column you can pass
	none_support=True to dw.add, or {colname: (coltype, True)} to the
	constructor. If you pass a DatasetColumn (from ds.columns[name]) you
	will inherit both type and None-support of that column. In dw.add the
	none_support argument takes precedence over (tuple/DatasetColumn in)
	the coltype argument.

	If you set hashlabel you can use dw.hashcheck(v) to check if v
	belongs in this slice. You can also call enable_hash_discard
	(in each slice, or after each set_slice), then the writer will
	discard anything that does not belong in this slice.

	If you are not in analysis and you wish to use the functions above
	you need to call dw.set_slice(sliceno) first.

	If you do not, you can instead get one of the splitting writer
	functions, that select which slice to use based on hashlabel, or
	round robin if there is no hashlabel.

	dw.get_split_write_dict()({column: value})
	dw.get_split_write_list()([value, value, ...])
	dw.get_split_write()(value, value, ...)

	These should of course be assigned to a local name for performance.

	It is permitted (but probably useless) to mix different write or
	split functions, but you can only use either write functions or
	split functions.

	You can also use dw.writers[colname] to get a typed_writer and use
	it as you please. The one belonging to the hashlabel will be
	filtering, and returns True if this is the right slice.

	If you need to handle everything yourself, set meta_only=True and
	use dw.column_filename(colname) to find the right files to write to.
	In this case you also need to call dw.set_lines(sliceno, count)
	before finishing. You should also call
	dw.set_minmax(sliceno, {colname: (min, max)}) if you can.

	If you are just copying from another dataset you can set copy_mode
	both here and in the iterator for that dataset for faster copying.
	"""

	_split = _split_dict = _split_list = _allwriters_ = None

	def __new__(cls, columns={}, filename=None, hashlabel=None, hashlabel_override=False, caption=None, previous=None, name='default', parent=None, meta_only=False, for_single_slice=None, copy_mode=False):
		"""columns can be {'name': 'type'} or {'name': ('type', none_support)}.
		It can also be {'name': DatasetColumn} to simplify basing your dataset on another."""
		name = _namechk(name)
		from accelerator.g import running
		if running == 'analysis':
			if name not in _datasetwriters:
				raise DatasetUsageError('Dataset with name "%s" not created' % (name,))
			if columns or filename or hashlabel or hashlabel_override or caption or previous or parent or meta_only or for_single_slice is not None:
				raise DatasetUsageError("Don't specify any arguments (except optionally name) in analysis")
			return _datasetwriters[name]
		else:
			if name in _datasetwriters:
				raise DatasetUsageError('Duplicate dataset name "%s"' % (name,))
			os.mkdir(name)
			obj = object.__new__(cls)
			obj._running = running
			obj.filename = uni(filename)
			obj.hashlabel = uni(hashlabel)
			obj.hashlabel_override = hashlabel_override
			obj.caption = uni(caption)
			obj.previous = _dsid(previous)
			obj.name = name
			obj.columns = {}
			obj.meta_only = meta_only
			obj._for_single_slice = for_single_slice
			obj._copy_mode = copy_mode
			obj._filenames = {}
			obj._fngen = _fngen()
			discard_columns = {k for k, v in columns.items() if v is None}
			columns = {k: v for k, v in columns.items() if v is not None}
			if parent:
				obj.parent = Dataset(parent)
				if not hashlabel_override:
					if obj.hashlabel is not None:
						if obj.hashlabel != obj.parent.hashlabel:
							raise DatasetUsageError("Hashlabel mismatch %r != %r" % (obj.hashlabel, obj.parent.hashlabel,))
					elif obj.parent.hashlabel in columns:
						obj.hashlabel = obj.parent.hashlabel
				parent_cols = obj.parent.columns
				unknown_discards = discard_columns - set(parent_cols)
				if unknown_discards:
					raise DatasetUsageError("Can't discard non-existant columns %r" % (unknown_discards,))
				obj._column_filter = set(parent_cols) - discard_columns
			else:
				if discard_columns:
					raise DatasetUsageError("Can't discard columns without a parent")
				obj.parent = None
			obj._started = False
			obj._lens = {}
			obj._minmax = {}
			obj._order = []
			for k, v in sorted(columns.items()):
				if v is None:
					continue
				obj.add(k, v)
			_datasetwriters[name] = obj
			return obj

	def add(self, colname, coltype, default=_nodefault, none_support=_nodefault):
		from accelerator.g import running
		if running != self._running:
			raise DatasetUsageError("Add all columns in the same step as creation")
		if self._started:
			raise DatasetUsageError("Add all columns before setting slice")
		if isinstance(coltype, tuple):
			if hasattr(coltype, 'type'):
				coltype, none_support_fb = coltype.type, coltype.none_support
			else:
				coltype, none_support_fb = coltype
		else:
			none_support_fb = False
		if none_support is _nodefault:
			none_support = none_support_fb
		else:
			none_support = bool(none_support)
		colname = uni(colname)
		coltype = uni(coltype)
		if colname in self.columns:
			raise DatasetUsageError("Column %s already exists" % (colname,))
		try:
			typed_writer(coltype) # gives error for unknown types
		except ValueError as e:
			raise DatasetUsageError(str(e))
		if none_support and coltype.startswith('bits'):
			raise DatasetUsageError("%s columns can't have None support" % (coltype,))
		self.columns[colname] = (coltype, default, none_support)
		self._order.append(colname)
		self._filenames[colname] = next(self._fngen)

	# These will be overwritten by set_slice
	def write(self, *a, **kw):
		raise DatasetUsageError("Call .set_slice(sliceno) before writing.")
	write_list = write
	write_dict = write

	def set_slice(self, sliceno):
		from accelerator import g
		if g.running == 'analysis' and self._for_single_slice != g.sliceno:
			if self._for_single_slice is not None:
				raise DatasetUsageError("This writer is for slice %d" % (self._for_single_slice,))
			else:
				raise DatasetUsageError("Only use set_slice in analysis together with for_single_slice")
		self._set_slice(sliceno)

	def _set_slice(self, sliceno):
		from accelerator.g import slices
		if self._started == 2:
			raise DatasetUsageError("Don't use both set_slice and a split writer")
		if not isinstance(sliceno, int_types) or sliceno < 0 or sliceno >= slices:
			raise DatasetUsageError("sliceno must be int in range(%d)" % (slices,))
		self.close()
		self.sliceno = sliceno
		writers = self._mkwriters(sliceno)
		if not self.meta_only:
			self.writers = writers
			self._mkwritefuncs()

	def column_filename(self, colname, sliceno=None):
		if sliceno is None:
			sliceno = self.sliceno
		return '%s/%d.%s' % (self.name, sliceno, self._filenames[colname],)

	def enable_hash_discard(self):
		"""Make the write functions silently discard data that does not
		hash to the current slice."""
		if self.hashlabel is None:
			raise DatasetUsageError("Can't enable hash discard without hashlabel")
		if self._started != 1:
			raise DatasetUsageError("Call enable_hash_discard after set_slice")
		self._mkwritefuncs(discard=True)

	def _mkwriters(self, sliceno, filtered=True):
		if not self.columns:
			raise DatasetUsageError("No columns in dataset")
		if self.hashlabel is not None and self.hashlabel not in self.columns:
			raise DatasetUsageError("Hashed column (%r) missing" % (self.hashlabel,))
		self._started = 2 - filtered
		if self.meta_only:
			return
		writers = {}
		for colname, (coltype, default, none_support) in self.columns.items():
			if self._copy_mode:
				coltype = _copy_mode_overrides.get(coltype, coltype)
			wt = typed_writer(coltype)
			kw = {'none_support': none_support} if default is _nodefault else {'default': default, 'none_support': none_support}
			fn = self.column_filename(colname, sliceno)
			if filtered and colname == self.hashlabel:
				from accelerator.g import slices
				w = wt(fn, hashfilter=(sliceno, slices), **kw)
				self.hashcheck = w.hashcheck
			else:
				w = wt(fn, **kw)
			writers[colname] = w
		return writers

	def _mkwritefuncs(self, discard=False):
		hl = self.hashlabel
		w_l = [self.writers[c].write for c in self._order]
		w = {k: w.write for k, w in self.writers.items()}
		wrong_slice_msg = "Attempted to write data for wrong slice"
		if hl is not None:
			hw = w.pop(hl)
			w_i = w.items()
			def write_dict(values):
				if hw(values[hl]):
					for k, w in w_i:
						w(values[k])
				elif not discard:
					raise DatasetUsageError(wrong_slice_msg)
			self.write_dict = write_dict
			hix = self._order.index(hl)
		else:
			w_i = w.items()
			def write_dict(values):
				for k, w in w_i:
					w(values[k])
			self.write_dict = write_dict
			hix = -1
		used_names = set()
		names = [_clean_name(n, used_names) for n in self._order]
		w_names = [_clean_name('w%d' % (ix,), used_names) for ix in range(len(w_l))]
		w_d = dict(zip(w_names, w_l))
		errcls = _clean_name('DatasetUsageError', used_names)
		w_d[errcls] = DatasetUsageError
		f = ['def write(' + ', '.join(names) + '):']
		f_list = ['def write_list(values):']
		if len(names) == 1: # only the hashlabel, no check needed
			f.append(' %s(%s)' % (w_names[0], names[0],))
			f_list.append(' %s(values[0])' % (w_names[0],))
		else:
			if hl is not None:
				f.append(' if %s(%s):' % (w_names[hix], names[hix],))
				f_list.append(' if %s(values[%d]):' % (w_names[hix], hix,))
			for ix in range(len(names)):
				if ix != hix:
					f.append('  %s(%s)' % (w_names[ix], names[ix],))
					f_list.append('  %s(values[%d])' % (w_names[ix], ix,))
			if hl is not None and not discard:
				f.append(' else: raise %s(%r)' % (errcls, wrong_slice_msg,))
				f_list.append(' else: raise %s(%r)' % (errcls, wrong_slice_msg,))
		eval(compile('\n'.join(f), '<DatasetWriter generated write>', 'exec'), w_d)
		self.write = w_d['write']
		eval(compile('\n'.join(f_list), '<DatasetWriter generated write_list>', 'exec'), w_d)
		self.write_list = w_d['write_list']

	@property
	def _allwriters(self):
		if self._allwriters_:
			return self._allwriters_
		from accelerator.g import slices
		self._allwriters_ = [self._mkwriters(sliceno, False) for sliceno in range(slices)]
		return self._allwriters_

	def get_split_write(self):
		return self._split or self._mksplit()['split']

	def get_split_write_list(self):
		return self._split_list or self._mksplit()['split_list']

	def get_split_write_dict(self):
		return self._split_dict or self._mksplit()['split_dict']

	def _mksplit(self):
		from accelerator import g
		if g.running == 'analysis' and self._for_single_slice != g.sliceno:
			if self._for_single_slice is not None:
				raise DatasetUsageError("This writer is for slice %d" % (self._for_single_slice,))
			else:
				raise DatasetUsageError("Only use a split writer in analysis together with for_single_slice")
		if self._started == 1:
			raise DatasetUsageError("Don't use both a split writer and set_slice")
		if self.parent and self.parent.hashlabel is not None and self.hashlabel is None:
			raise DatasetUsageError("Can't use a split writer on hashed dataset when not writing the hash column.")
		used_names = set()
		names = [_clean_name(n, used_names) for n in self._order]
		def key(t):
			return self._order.index(t[0])
		def d2l(d):
			return [w.write for _, w in sorted(d.items(), key=key)]
		f_____ = ['def split(' + ', '.join(names) + '):']
		f_list = ['def split_list(v):']
		f_dict = ['def split_dict(d):']
		from accelerator.g import slices
		hl = self.hashlabel
		name_w_l = _clean_name('w_l', used_names)
		name_hsh = _clean_name('hsh', used_names)
		name_cyc = _clean_name('cyc', used_names)
		name_next = _clean_name('next', used_names)
		name_writers = _clean_name('writers', used_names)
		w_d = {}
		w_d[name_writers] = [d2l(d) for d in self._allwriters]
		w_d[name_next] = next
		if hl is not None:
			w_d[name_hsh] = self._allwriters[0][hl].hash
			prefix = '%s = %s[%s(' % (name_w_l, name_writers, name_hsh,)
			hix = self._order.index(hl)
			f_____.append('%s%s) %% %d]' % (prefix, names[hix], slices,))
			f_list.append('%sv[%d]) %% %d]' % (prefix, hix, slices,))
			f_dict.append('%sd[%r]) %% %d]' % (prefix, hl, slices,))
		else:
			from itertools import cycle
			w_d[name_cyc] = cycle(range(slices))
			code = '%s = %s[%s(%s)]' % (name_w_l, name_writers, name_next, name_cyc,)
			f_____.append(code)
			f_list.append(code)
			f_dict.append(code)
		for ix in range(len(names)):
			f_____.append('%s[%d](%s)' % (name_w_l, ix, names[ix],))
			f_list.append('%s[%d](v[%d])' % (name_w_l, ix, ix,))
			f_dict.append('%s[%d](d[%r])' % (name_w_l, ix, self._order[ix],))
		eval(compile('\n '.join(f_____), '<DatasetWriter generated split_write>'     , 'exec'), w_d)
		eval(compile('\n '.join(f_list), '<DatasetWriter generated split_write_list>', 'exec'), w_d)
		eval(compile('\n '.join(f_dict), '<DatasetWriter generated split_write_dict>', 'exec'), w_d)
		self._split = w_d['split']
		self._split_list = w_d['split_list']
		self._split_dict = w_d['split_dict']
		return w_d

	def _close(self, sliceno, writers):
		lens = {}
		minmax = {}
		for k, w in writers.items():
			lens[k] = w.count
			minmax[k] = (w.min, w.max,)
			w.close()
		len_set = set(lens.values())
		if len(len_set) != 1:
			raise DatasetUsageError("Not all columns have the same linecount in slice %d: %r" % (sliceno, lens))
		self._lens[sliceno] = len_set.pop()
		self._minmax[sliceno] = minmax

	def close(self):
		if self._started == 2:
			for sliceno, writers in enumerate(self._allwriters):
				self._close(sliceno, writers)
		else:
			if hasattr(self, 'writers'):
				self._close(self.sliceno, self.writers)
				del self.writers

	def discard(self):
		del _datasetwriters[self.name]
		from shutil import rmtree
		rmtree(self.name)

	def set_lines(self, sliceno, count):
		if not self.meta_only:
			raise DatasetUsageError("Don't try to set lines for writers that actually write")
		self._lens[sliceno] = count

	def set_minmax(self, sliceno, minmax):
		if not self.meta_only:
			raise DatasetUsageError("Don't try to set minmax for writers that actually write")
		self._minmax[sliceno] = minmax

	def finish(self):
		"""Normally you don't need to call this, but if you want to
		pass yourself as a dataset to a subjob you need to call
		this first."""
		from accelerator.g import running, slices, job
		if running != self._running and running != 'synthesis':
			raise DatasetUsageError("Finish where you started or in synthesis")
		self.close()
		if len(self._lens) != slices:
			raise DatasetUsageError("Not all slices written, missing %r" % (set(range(slices)) - set(self._lens),))
		args = dict(
			columns={k: (v[0].split(':')[-1], v[2]) for k, v in self.columns.items()},
			filenames=self._filenames,
			lines=self._lens,
			minmax=self._minmax,
			filename=self.filename,
			hashlabel=self.hashlabel,
			caption=self.caption,
			previous=self.previous,
			name=self.name,
		)
		if self.parent:
			self.parent.append(hashlabel_override=self.hashlabel_override, column_filter=self._column_filter, **args)
			res = Dataset((job, self.name))
		else:
			res = Dataset.new(**args)
		del _datasetwriters[self.name]
		_datasets_written.append(self.name)
		return res

class DatasetChain(_ListTypePreserver):
	"""
	These are lists of datasets returned from Dataset.chain.
	They exist to provide some convenience methods on chains.
	"""

	def _minmax(self, column, minmax):
		vl = []
		nan_seen = None
		for ds in self:
			c = ds.columns.get(column)
			if c:
				v = getattr(c, minmax)
				if isinstance(v, float) and isnan(v):
					nan_seen = v
				elif v is not None:
					vl.append(v)
		if vl:
			return getattr(builtins, minmax)(vl)
		else:
			return nan_seen

	def min(self, column):
		"""Min value for column over the whole chain.
		Will be None if no dataset in the chain contains column,
		if all datasets are empty or if column has a type without
		min/max tracking"""
		return self._minmax(column, 'min')

	def max(self, column):
		"""Max value for column over the whole chain.
		Will be None if no dataset in the chain contains column,
		if all datasets are empty or if column has a type without
		min/max tracking"""
		return self._minmax(column, 'max')

	def lines(self, sliceno=None):
		"""Number of rows in this chain, optionally for a specific slice."""
		if sliceno is None:
			sel = sum
		else:
			sel = itemgetter(sliceno)
		return sum(sel(ds.lines) for ds in self)

	def column_counts(self):
		"""Counter {colname: occurances}"""
		from itertools import chain
		return Counter(chain.from_iterable(ds.columns.keys() for ds in self))

	def column_count(self, column):
		"""How many datasets in this chain contain column"""
		return sum(column in ds.columns for ds in self)

	def with_column(self, column):
		"""Chain without any datasets that don't contain column"""
		return self.__class__(ds for ds in self if column in ds.columns)

	def none_support(self, column):
		"""If any dataset in the chain has None support for this column"""
		return True in (ds.columns[column].none_support for ds in self if column in ds.columns)

	def iterate(self, sliceno, columns=None, range=None, sloppy_range=False, hashlabel=None, pre_callback=None, post_callback=None, filters=None, translators=None, status_reporting=True, rehash=False, slice=None, copy_mode=False):
		"""Iterate the datasets in this chain. See Dataset.iterate_list for usage"""
		return Dataset.iterate_list(sliceno, columns, self, range=range, sloppy_range=sloppy_range, hashlabel=hashlabel, pre_callback=pre_callback, post_callback=post_callback, filters=filters, translators=translators, status_reporting=status_reporting, rehash=rehash, slice=slice, copy_mode=copy_mode)


def range_check_function(bottom, top):
	"""Returns a function that checks if bottom <= arg < top, allowing bottom and/or top to be None"""
	import operator
	if top is None:
		if bottom is None:
			# Can't currently happen (checked before calling this), but let's do something reasonable
			return lambda _: True
		else:
			return partial(operator.le, bottom)
	elif bottom is None:
		return partial(operator.gt, top)
	else:
		def range_f(v):
			return v >= bottom and v < top
		return range_f

class SkipDataset(Exception):
	"""Raise this in pre_callback to skip iterating the coming dataset
	(or the remaining slices of it)"""

class SkipSlice(Exception):
	"""Raise this in pre_callback to skip iterating the coming slice
	(if your callback doesn't want sliceno, this is the same as SkipDataset)"""

def job_datasets(job):
	"""All datasets in a job"""
	if isinstance(job, Dataset):
		job = job.job
	else:
		job = Job(job)
	fn = job.filename('datasets.txt')
	if not os.path.exists(fn):
		# It's not an error to list datasets in a job without them.
		return []
	with open(fn, 'r', encoding='utf-8') as fh:
		names = [line[:-1] for line in fh]
	res = []
	# Do this backwards to improve chances that we take advantage of cache.
	# (Names are written to datasets.txt in finish() order.)
	for name in reversed(names):
		res.append(Dataset(job, name))
	return list(reversed(res))



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Exercise different valid and invalid parent/hashlabel combinations in
DatasetWriter.
'''

from accelerator import DatasetUsageError, NoSuchDatasetError

def synthesis(slices, job):
	parent = job.datasetwriter(name='parent', columns={'a': 'ascii'}, hashlabel='a')
	w = parent.get_split_write()
	w('a')
	w('aa')
	w('aaa')
	parent = parent.finish()
	def test_good(name, specified_parent):
		dw = job.datasetwriter(name=name, columns={'b': 'ascii'}, parent=specified_parent)
		for sliceno in range(slices):
			dw.set_slice(sliceno)
			for v in parent.iterate(sliceno, 'a'):
				dw.write(v + v)
		ds = dw.finish()
		assert set(ds.iterate(None)) == {('a', 'aa'), ('aa', 'aaaa'), ('aaa', 'aaaaaa')}
	test_good('ok', parent)
	test_good('also ok', (job, 'parent')) # different way of specifying parent
	dw = job.datasetwriter(columns={'b': 'ascii'}, parent=parent)
	try:
		dw.get_split_write()
		raise Exception('get_split_write allowed when not writing hash label')
	except DatasetUsageError:
		pass
	dw.discard()
	try:
		dw = job.datasetwriter(name='fail', columns={'b': 'ascii'}, parent=parent, hashlabel='b')
		raise Exception('Allowed changing hashlabel without hashlabel_override')
	except DatasetUsageError:
		pass
	# This is permitted, though kind of questionable since we pretty much have to write the same thing in b as in a
	dw = job.datasetwriter(name='overridden', columns={'b': 'ascii'}, parent=parent, hashlabel='b', hashlabel_override=True)
	w = dw.get_split_write()
	w('a')
	w('aa')
	w('aaa')
	ds = dw.finish()
	assert set(ds.iterate(None)) == {('a', 'a'), ('aa', 'aa'), ('aaa', 'aaa')}
	dw = job.datasetwriter(columns={'b': 'ascii'}, parent=parent, hashlabel='a')
	try:
		dw.set_slice(0)
		raise Exception('set_slice allowed when not writing hash label')
	except DatasetUsageError:
		pass
	dw.discard()
	dup = job.datasetwriter(name='dup')
	try:
		job.datasetwriter(name='dup')
		raise Exception('dataset name reuse allowed')
	except DatasetUsageError:
		pass
	dup.discard()
	try:
		job.datasetwriter(name='parent')
		raise Exception('already finished dataset name reuse allowed')
	except OSError:
		pass
	try:
		job.datasetwriter(parent=(job, 'nah'))
		raise Exception('allowed non-existant parent')
	except NoSuchDatasetError:
		pass



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
# Modifications copyright (c) 2019-2020 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

import os
import re

from collections import namedtuple
from functools import wraps

from accelerator.compat import unicode, PY2, PY3, open, iteritems
from accelerator.error import NoSuchJobError, NoSuchWorkdirError


# WORKDIRS should live in the Automata class, but only for callers
# (methods read it too, though hopefully only through the functions in this module)

WORKDIRS = {}


def dirnamematcher(name):
	return re.compile(re.escape(name) + r'-[0-9]+$').match


def _cachedprop(meth):
	@property
	@wraps(meth)
	def wrapper(self):
		if meth.__name__ not in self._cache:
			self._cache[meth.__name__] = meth(self)
		return self._cache[meth.__name__]
	return wrapper

class Job(unicode):
	"""
	A string that is a jobid, but also has some extra properties:
	.method The job method (can be the "name" when from build or urd).
	.number The job number as an int.
	.workdir The workdir name (the part before -number in the jobid)
	.path The filesystem directory where the job is stored.
	.params setup.json from this job.
	.post post.json from this job.
	.datasets list of Datasets in this job.
	And some functions:
	.withfile a JobWithFile with this job.
	.filename to join .path with a filename.
	.load to load a pickle.
	.json_load to load a json file.
	.open to open a file (like standard open)
	.dataset to get a named Dataset.
	.output to get what the job printed.
	.link_result to put a link in result_directory that points to a file in this job.

	Decays to a (unicode) string when pickled.
	"""
	def __new__(cls, jobid, method=None):
		obj = unicode.__new__(cls, jobid)
		try:
			obj.workdir, tmp = jobid.rsplit('-', 1)
			obj.number = int(tmp)
		except ValueError:
			raise NoSuchJobError('Not a valid jobid: "%s".' % (jobid,))
		obj._cache = {}
		if method:
			obj._cache['method'] = method
		return obj

	@classmethod
	def _create(cls, name, number):
		return Job('%s-%d' % (name, number,))

	@_cachedprop
	def method(self):
		return self.params.method

	@property
	def path(self):
		if self.workdir not in WORKDIRS:
			raise NoSuchWorkdirError('Not a valid workdir: "%s"' % (self.workdir,))
		return os.path.join(WORKDIRS[self.workdir], self)

	def filename(self, filename, sliceno=None):
		if sliceno is not None:
			filename = '%s.%d' % (filename, sliceno,)
		return os.path.join(self.path, filename)

	def open(self, filename, mode='r', sliceno=None, encoding=None, errors=None):
		assert 'r' in mode, "Don't write to other jobs"
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		return open(self.filename(filename, sliceno), mode, encoding=encoding, errors=errors)

	def files(self, pattern='*'):
		from fnmatch import filter
		return set(filter(self.post.files, pattern))

	def withfile(self, filename, sliced=False, extra=None):
		return JobWithFile(self, filename, sliced, extra)

	@_cachedprop
	def params(self):
		from accelerator.extras import job_params
		return job_params(self)

	@_cachedprop
	def post(self):
		from accelerator.extras import job_post
		return job_post(self)

	def load(self, filename='result.pickle', sliceno=None, encoding='bytes'):
		"""blob.load from this job"""
		from accelerator.extras import pickle_load
		return pickle_load(self.filename(filename, sliceno), encoding=encoding)

	def json_load(self, filename='result.json', sliceno=None, unicode_as_utf8bytes=PY2):
		from accelerator.extras import json_load
		return json_load(self.filename(filename, sliceno), unicode_as_utf8bytes=unicode_as_utf8bytes)

	def dataset(self, name='default'):
		from accelerator.dataset import Dataset
		return Dataset(self, name)

	@_cachedprop
	def datasets(self):
		from accelerator.dataset import job_datasets
		return job_datasets(self)

	def output(self, what=None):
		if isinstance(what, int):
			fns = [str(what)]
		else:
			assert what in (None, 'prepare', 'analysis', 'synthesis'), 'Unknown output %r' % (what,)
			if what in (None, 'analysis'):
				fns = [str(sliceno) for sliceno in range(self.params.slices)]
				if what is None:
					fns = ['prepare'] + fns + ['synthesis']
			else:
				fns = [what]
		res = []
		for fn in fns:
			fn = self.filename('OUTPUT/' + fn)
			if os.path.exists(fn):
				with open(fn, 'rt', encoding='utf-8', errors='backslashreplace') as fh:
					res.append(fh.read())
		return ''.join(res)

	def link_result(self, filename='result.pickle', linkname=None):
		"""Put a symlink to filename in result_directory
		Only use this in a build script."""
		from accelerator.g import running
		assert running == 'build', "Only link_result from a build script"
		from accelerator.shell import cfg
		if linkname is None:
			linkname = filename
		result_directory = cfg['result_directory']
		dest_fn = os.path.join(result_directory, linkname)
		try:
			os.remove(dest_fn + '_')
		except OSError:
			pass
		source_fn = os.path.join(self.path, filename)
		assert os.path.exists(source_fn), "Filename \"%s\" does not exist in jobdir \"%s\"!" % (filename, self.path)
		os.symlink(source_fn, dest_fn + '_')
		os.rename(dest_fn + '_', dest_fn)

	def chain(self, length=-1, reverse=False, stop_job=None):
		"""Like Dataset.chain but for jobs."""
		if isinstance(stop_job, dict):
			assert len(stop_job) == 1, "Only pass a single stop_job={job: name}"
			stop_job, stop_name = next(iteritems(stop_job))
			if stop_job:
				stop_job = Job(stop_job).params.jobs.get(stop_name)
		chain = []
		current = self
		while length != len(chain) and current and current != stop_job:
			chain.append(current)
			current = current.params.jobs.get('previous')
		if not reverse:
			chain.reverse()
		return chain

	# Look like a string after pickling
	def __reduce__(self):
		return unicode, (unicode(self),)


class CurrentJob(Job):
	"""The currently running job (as passed to the method),
	with extra functions for writing data."""

	def __new__(cls, jobid, params, result_directory, input_directory):
		obj = Job.__new__(cls, jobid, params.method)
		obj._cache['params'] = params
		obj.result_directory = result_directory
		obj.input_directory = input_directory
		return obj

	def save(self, obj, filename='result.pickle', sliceno=None, temp=None):
		from accelerator.extras import pickle_save
		pickle_save(obj, filename, sliceno, temp=temp)

	def json_save(self, obj, filename='result.json', sliceno=None, sort_keys=True, temp=None):
		from accelerator.extras import json_save
		json_save(obj, filename, sliceno, sort_keys=sort_keys, temp=temp)

	def datasetwriter(self, columns={}, filename=None, hashlabel=None, hashlabel_override=False, caption=None, previous=None, name='default', parent=None, meta_only=False, for_single_slice=None, copy_mode=False):
		from accelerator.dataset import DatasetWriter
		return DatasetWriter(columns=columns, filename=filename, hashlabel=hashlabel, hashlabel_override=hashlabel_override, caption=caption, previous=previous, name=name, parent=parent, meta_only=meta_only, for_single_slice=for_single_slice, copy_mode=copy_mode)

	def open(self, filename, mode='r', sliceno=None, encoding=None, errors=None, temp=None):
		"""Mostly like standard open with sliceno and temp,
		but you must use it as context manager
		with job.open(...) as fh:
		and the file will have a temp name until the with block ends.
		"""
		if 'r' in mode:
			return Job.open(self, filename, mode, sliceno, encoding, errors)
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		if PY3 and 'x' not in mode:
			mode = mode.replace('w', 'x')
		def _open(fn, _mode):
			# ignore the passed mode, use the one we have
			return open(fn, mode, encoding=encoding, errors=errors)
		from accelerator.extras import FileWriteMove
		fwm = FileWriteMove(self.filename(filename, sliceno), temp=temp)
		fwm._open = _open
		return fwm

	def register_file(self, filename):
		"""Record a file produced by this job. Normally you would use
		job.open to have this happen automatically, but if the file was
		produced in a way where that is not practical you can use this
		to register it."""
		filename = self.filename(filename)
		assert os.path.exists(filename)
		from accelerator.extras import saved_files
		saved_files[filename] = 0

	def input_filename(self, filename):
		return os.path.join(self.input_directory, filename)

	def open_input(self, filename, mode='r', encoding=None, errors=None):
		assert 'r' in mode, "Don't write to input files"
		if 'b' not in mode and encoding is None:
			encoding = 'utf-8'
		return open(self.input_filename(filename), mode, encoding=encoding, errors=errors)

class JobWithFile(namedtuple('JobWithFile', 'job name sliced extra')):
	def __new__(cls, job, name, sliced=False, extra=None):
		assert not name.startswith('/'), "Specify relative filenames to JobWithFile"
		return tuple.__new__(cls, (Job(job), name, bool(sliced), extra,))

	def filename(self, sliceno=None):
		if sliceno is None:
			assert not self.sliced, "A sliced file requires a sliceno"
		else:
			assert self.sliced, "An unsliced file can not have a sliceno"
		return self.job.filename(self.name, sliceno)

	def load(self, sliceno=None, encoding='bytes'):
		"""blob.load this file"""
		from accelerator.extras import pickle_load
		return pickle_load(self.filename(sliceno), encoding=encoding)

	def json_load(self, sliceno=None, unicode_as_utf8bytes=PY2):
		from accelerator.extras import json_load
		return json_load(self.filename(sliceno), unicode_as_utf8bytes=unicode_as_utf8bytes)

	def open(self, mode='r', sliceno=None, encoding=None, errors=None):
		return self.job.open(self.name, mode, sliceno, encoding, errors)



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
# Modifications copyright (c) 2021 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from accelerator.compat import terminal_size
from accelerator.unixhttp import call
from accelerator.shell import printdesc
from collections import defaultdict

def main(argv, cfg):
	prog = argv.pop(0)
	if '--help' in argv or '-h' in argv:
		print('usage: %s [method]' % (prog,))
		print('gives description and options for method,')
		print('or lists methods with no method specified.')
		return
	methods = call(cfg.url + '/methods')
	columns = terminal_size().columns
	if argv:
		for name in argv:
			if name in methods:
				data = methods[name]
				print('%s.%s:' % (data.package, name,))
				if data.description.text:
					for line in data.description.text.split('\n'):
						if line:
							print(' ', line)
						else:
							print()
					print()
				if cfg.get('interpreters'):
					print('Runs on <%s> %s' % (data.version, data.description.interpreter,))
					print()
				for k in ('datasets', 'jobs',):
					if data.description.get(k):
						print('%s:' % (k,))
						klen = max(len(k) for k in data.description[k])
						template = '  %%-%ds # %%s' % (klen,)
						for k, v in data.description[k].items():
							if v:
								print(template % (k, v[0],))
								for cmt in v[1:]:
									print(template % ('', cmt,))
							else:
								print(' ', k)
				if data.description.get('options'):
					print('options:')
					klen = max(len(k) for k in data.description.options)
					vlens = [len(v[0]) for v in data.description.options.values() if len(v) > 1]
					vlen = max(vlens or [0])
					firstlen = klen + vlen + 5
					template = '  %%-%ds = %%s' % (klen,)
					template_cmt = '%%-%ds  # %%s' % (firstlen,)
					for k, v in data.description.options.items():
						first = template % (k, v[0],)
						if len(v) > 1:
							afterlen = max(len(cmt) for cmt in v[1:]) + firstlen + 4
							if afterlen <= columns:
								print(template_cmt % (first, v[1],))
								for cmt in v[2:]:
									print(template_cmt % ('', cmt,))
							else:
								for cmt in v[1:]:
									print('  #', cmt)
								print(first)
						else:
							print(first)
			else:
				print('Method %r not found' % (name,))
	else:
		by_package = defaultdict(list)
		for name, data in sorted(methods.items()):
			by_package[data.package].append(name)
		by_package.pop('accelerator.test_methods', None)
		for package, names in sorted(by_package.items()):
			print('%s:' % (package,))
			for name in names:
				printdesc(name, methods[name].description.text.split('\n')[0], columns)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from collections import defaultdict
from operator import attrgetter
from collections import namedtuple

from accelerator.compat import iteritems, itervalues

from accelerator.extras import _job_params, job_post, OptionEnum, OptionDefault


Job = namedtuple('Job', 'id method optset hash time total')

_control = None # control.Main instance, global for use in _mkjob, set when DataBase is initialized.

def _mkjob(setup):
	# Start with defaults, update with actual options
	def optfilter(d):
		res = {}
		for k, v in iteritems(d):
			if isinstance(v, OptionEnum):
				v = None
			elif isinstance(v, OptionDefault):
				v = v.default
			res[k] = v
		return res
	if setup.method in _control.Methods.params:
		d = {k: optfilter(v) for k, v in iteritems(_control.Methods.params[setup.method].defaults)}
	else:
		d = {}
	for k, v in iteritems(d):
		v.update(setup[k])
	optset = _control.Methods.params2optset({setup.method: d})
	job = Job(
		id     = setup.jobid,
		method = setup.method,
		optset = optset,
		hash   = setup.hash,
		time   = setup.starttime,
		total  = setup.exectime.total,
	)
	return job

def _mklistinfo(setup):
	return dict(
		method=setup.method,
		totaltime=setup.exectime.total,
		current=True,
	)

def _get_params(jobid):
	try:
		return jobid, (_job_params(jobid), list(job_post(jobid).subjobs))
	except:
		from traceback import print_exc
		print_exc()
		raise

class _ParamsDict(defaultdict):
	def __missing__(self, key):
		self[key] = value = _get_params(key)[1]
		return value
_paramsdict = _ParamsDict()

class DataBase:
	def __init__(self, control):
		global _control
		assert not _control, "Only one DataBase instance allowed"
		_control = control

	def _update_begin(self):
		self._fsjid = set()

	def add_single_jobid(self, jobid):
		setup = _paramsdict[jobid][0]
		job = _mkjob(setup)
		self.db_by_method[job.method].insert(0, job)
		self.db_by_workdir[job.id.rsplit('-', 1)[0]][job.id] = _mklistinfo(setup)
		return job

	def _update_workspace(self, WorkSpace, pool, verbose=False):
		"""Insert all items in WorkSpace in database (call update_finish too)"""
		if verbose:
			print("DATABASE:  update for \"%s\"" % WorkSpace.name)
		filesystem_jobids = WorkSpace.valid_jobids
		self._fsjid.update(filesystem_jobids)
		if verbose > 1:
			print('DATABASE:  update found these jobids in workdir', filesystem_jobids)
		# Insert any new jobids, including with invalid hash
		new_jobids = filesystem_jobids.difference(_paramsdict)
		if new_jobids:
			_paramsdict.update(pool.imap_unordered(_get_params, new_jobids, chunksize=64))
		if verbose:
			print("DATABASE:  Database \"%s\" contains %d potential items" % (WorkSpace.name, len(filesystem_jobids), ))

	def _update_finish(self, dict_of_hashes, verbose=False):
		"""Filters in-use database on valid hashes.
		Always call after (a sequence of) update_workspace calls.
		"""
		# discard cached setup.json from any gone jobs
		# (so we reload it if they reappear, and also so we don't see them here)
		for j in set(_paramsdict) - self._fsjid:
			del _paramsdict[j]
		discarded_due_to_hash_list = []
		self.db_by_workdir = defaultdict(dict) # includes all known jobs, not just current ones.

		# Keep only jobs with valid hashes.
		job_candidates = {}
		for setup, subjobs in itervalues(_paramsdict):
			if setup.hash in dict_of_hashes.get(setup.method, ()):
				job_candidates[setup.jobid] = (setup, subjobs)
			else:
				discarded_due_to_hash_list.append(setup.jobid)
			self.db_by_workdir[setup.jobid.rsplit('-', 1)[0]][setup.jobid] = _mklistinfo(setup)

		# Keep only jobs where all subjobs are kept.
		discarded_due_to_subjobs = []
		done = False
		while not done:
			done = True
			for setup, subjobs in list(itervalues(job_candidates)):
				for j in subjobs:
					if j not in job_candidates:
						done = False
						discarded_due_to_subjobs.append(setup.jobid)
						del job_candidates[setup.jobid]
						break

		for d in self.db_by_workdir.values():
			for jid, li in d.items():
				if jid not in job_candidates:
					li['current'] = False

		# Keep lists of jobs per method, only with valid hashes and subjobs.
		self.db_by_method = defaultdict(list)
		for setup, _ in itervalues(job_candidates):
			job = _mkjob(setup)
			self.db_by_method[job.method].append(job)
		# Newest first
		for l in itervalues(self.db_by_method):
			l.sort(key=attrgetter('time'), reverse=True)
		if verbose:
			if discarded_due_to_hash_list:
				print("DATABASE:  discarding due to unknown hash: %s" % ', '.join(discarded_due_to_hash_list))
			print("DATABASE:  Full database contains %d items" % (sum(len(v) for v in itervalues(self.db_by_method)),))

	def match_complex(self, reqlist):
		for method, uid, opttuple in reqlist:
			# These are already sorted newest to oldest.
			for job in self.db_by_method[method]:
				if opttuple.issubset(job.optset):
					yield uid, job
					break

	def match_exact(self, reqlist):
		for method, uid, opttuple in reqlist:
			# These are already sorted newest to oldest.
			for job in self.db_by_method[method]:
				if opttuple == job.optset:
					yield uid, job
					break



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test Dataset.merge() and the dataset_merge method.
'''

from accelerator.dataset import DatasetWriter, DatasetUsageError
from accelerator import subjobs
from accelerator.dispatch import JobError

def mkds(name, columns, data, **kw):
	columns = dict.fromkeys(columns, 'int32')
	dw = DatasetWriter(name=name, columns=columns, **kw)
	write = dw.get_split_write()
	for v in data:
		write(*v)
	return dw.finish()

merges = {}
def merge(name, a, b, **kw):
	merges[name] = ((a, b), kw)
	return a.merge(b, name=name, **kw)

failed_merges = []
def fail_merge(a, b, **kw):
	failed_merges.append(((a, b), kw,))
	try:
		a.merge(b, name='failme', **kw)
	except DatasetUsageError:
		return
	raise Exception("Merging %s and %s with %r didn't fail as it should have" % (a, b, kw,))

checks = {}
def check(ds, want):
	if '/' in ds:
		checks[ds.name] = want
	got = list(ds.iterate_chain(None))
	got.sort()
	assert got == want, "%s contained %r not %r as expetected" % (ds, got, want,)

def synthesis(params):
	a0 = mkds('a0', ['0', '1'], [(1, 2), (3, 4), (5, 6)])
	a1 = mkds('a1', ['0', '1'], [(7, 8), (9, 10), (11, 12)], previous=a0)
	b0 = mkds('b0', ['1', '2'], [(13, 14), (15, 16), (17, 18)], parent=a0)
	b1 = mkds('b1', ['1', '2'], [(19, 20), (21, 22), (23, 24)], parent=a1, previous=b0)
	c0 = mkds('c0', ['3'], [(25,), (26,), (27,)], parent=a0)
	c1 = mkds('c1', ['3'], [(28,), (29,), (30,)], parent=a1, previous=c0)
	# a contains columns 0 and 1
	# b contains columns 0, 1 and 2 with 0 from a
	# c contains columns 0, 1 and 3 with 0 and 1 from a
	bc0 = merge('bc0', b0, c0) # simple merge, one overlapping column
	# bc contains columns 0, 1, 2 and 3, with 0 and 1 from a (via c), 2 from b and 3 from c
	check(bc0, [(1, 2, 14, 25), (3, 4, 16, 26), (5, 6, 18, 27)])
	bc1 = merge('bc1', b1, c1, previous=bc0) # chained
	check(bc1, [(1, 2, 14, 25), (3, 4, 16, 26), (5, 6, 18, 27), (7, 8, 20, 28), (9, 10, 22, 29), (11, 12, 24, 30)])
	cb0 = merge('cb0', c0, b0) # other direction, getting the other "1" column.
	# cb contains columns 0, 1, 2 and 3, with 0 from a (via b), 1 and 2 from b and 3 from c
	check(cb0, [(1, 13, 14, 25), (3, 15, 16, 26), (5, 17, 18, 27)])
	d0 = mkds('d0', ['4'], [(37,), (38,), (39,)], parent=c0)
	bd0 = merge('bd0', b0, d0) # immediate parents are not shared
	# bd contains columns 0, 1, 2, 3 and 4, with 0 and 1 from a (via d -> c -> a), 2 from b, 3 from c and 4 from d
	check(bd0, [(1, 2, 14, 25, 37), (3, 4, 16, 26, 38), (5, 6, 18, 27, 39)])
	# more than two datasets with complex parent relationship
	# merged in two stages here, but a single dataset_merge job later.
	cbd0 = merge('cbd0', c0, bd0)
	del merges['cbd0']
	cbdb0 = merge('cbdb0', cbd0, b0)
	merges['cbdb0'] = ((c0, bd0, b0), {})
	# cbdb contains columns 0, 1, 2, 3 and 4, with 0 from a (via d -> c -> a), 1 and 2 from b, 3 from c and 4 from d
	check(cbdb0, [(1, 13, 14, 25, 37), (3, 15, 16, 26, 38), (5, 17, 18, 27, 39)])
	fail_merge(a0, a1) # no parents
	fail_merge(b0, b1) # parents not shared
	fail_merge(b0, b0) # merge with self
	other = mkds('other', ['5'], [(31,), (32,), (33,)])
	fail_merge(a0, other) # parents not shared
	aother = merge('aother', a0, other, allow_unrelated=True)
	# aother contains 0 and 1 from a, 5 from other
	check(aother, [(1, 2, 31), (3, 4, 32), (5, 6, 33)])

	# check hashed datasets too
	ab_a = mkds('ab_a', ['a', 'b'], [(1, 2), (3, 4), (5, 6)], hashlabel='a')
	ab_b = mkds('ab_b', ['a', 'b'], [(7, 8), (9, 10), (11, 12)], hashlabel='b')
	ac_a = mkds('ac_a', ['a', 'c'], [(1, 14), (3, 15), (5, 16)], hashlabel='a') # a values must match ab_a
	fail_merge(ab_a, ab_b, allow_unrelated=True) # different hashlabels
	abac_a = merge('abac_a', ab_a, ac_a, allow_unrelated=True)
	assert abac_a.hashlabel == 'a'
	check(abac_a, [(1, 2, 14), (3, 4, 15), (5, 6, 16)])

	# merge hashed with unhashed (which we align with the hashlabel manually)
	dw = DatasetWriter(name='d_none', columns={'d': 'number'})
	for sliceno in range(params.slices):
		dw.set_slice(sliceno)
		for v in ab_a.iterate(sliceno, 'a'):
			dw.write(v + 16)
	d_none = dw.finish()
	abd_a = merge('abd_a', ab_a, d_none, allow_unrelated=True)
	assert abd_a.hashlabel == 'a'
	check(abd_a, [(1, 2, 17), (3, 4, 19), (5, 6, 21)])
	# other way round should affect nothing here
	dab_a = merge('dab_a', ab_a, d_none, allow_unrelated=True)
	assert dab_a.hashlabel == 'a'
	check(dab_a, [(1, 2, 17), (3, 4, 19), (5, 6, 21)])

	# the same test but with the lines in the wrong slices:
	dw = DatasetWriter(name='e_none', columns={'e': 'number'})
	e_done = False
	for sliceno in range(params.slices):
		dw.set_slice(sliceno)
		# there are 3 lines in total, some slice will not have all of them.
		if ab_a.lines[sliceno] != 3 and not e_done:
			dw.write(17)
			dw.write(19)
			dw.write(21)
			e_done = True
	assert e_done
	e_none = dw.finish()
	fail_merge(ab_a, e_none, allow_unrelated=True)

	# and finally test all we tested above using the dataset_merge method too
	for name, (parents, kw) in merges.items():
		a_ds = dict(source=parents)
		if 'previous' in kw:
			a_ds['previous'] = kw.pop('previous')
		jid = subjobs.build('dataset_merge', datasets=a_ds, options=kw)
		check(jid.dataset(), checks[name])
	for parents, kw in failed_merges:
		try:
			subjobs.build('dataset_merge', datasets=dict(source=parents), options=kw)
		except JobError:
			continue
		raise Exception("dataset_merge incorrectly allowed %r with %r" % (parents, kw))



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Merge two or more datasets.
The datasets must have the same number of lines in each slice
and if they do not have a common ancestor you must set
allow_unrelated=True.

Columns from later datasets override columns of the same name
from earlier datasets.
'''

options = dict(
	allow_unrelated=False,
)
datasets = (['source'], 'previous',)

def synthesis():
	assert len(datasets.source) >= 2, 'Must have at least two datasets to join'
	current = datasets.source[0]
	for ix, ds in enumerate(datasets.source[1:-1]):
		current = current.merge(ds, name=str(ix), allow_unrelated=options.allow_unrelated)
	current.merge(datasets.source[-1], allow_unrelated=options.allow_unrelated, previous=datasets.previous)



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset iteration slicing.
'''

from accelerator.error import DatasetError

def synthesis(job, slices):
	dw = job.datasetwriter({'a': 'int32'}, name='first')
	dw.set_slice(0)
	dw.write(0)
	dw.write(1)
	dw.write(2)
	dw.set_slice(1)
	dw.set_slice(2)
	for ix in range(3, 100):
		dw.write(ix)
	for sliceno in range(3, slices):
		dw.set_slice(sliceno)
	ds = dw.finish()
	expect = list(range(100))
	def get(sliceno, slice, columns='a'):
		return list(ds.iterate(sliceno, columns, slice=slice))
	def get_chain(sliceno, slice, columns='a'):
		res = list(ds.iterate_chain(sliceno, columns, slice=slice))
		assert res == list(ds.chain().iterate(sliceno, columns, slice=slice))
		return res
	assert get(None, None) == expect
	assert get(None, -5) == expect[-5:]
	assert get(None, 97) == expect[97:]
	assert get(None, slice(93)) == expect[:93]
	assert get(None, slice(-97)) == expect[:-97]
	assert get(None, slice(0, -15)) == expect[:-15]
	assert get(None, slice(None, -15)) == expect[:-15]
	assert get(None, slice(10, 50, 3)) == expect[10:50:3]
	assert get(None, slice(None, None, 7)) == expect[::7]
	assert get(None, slice(1, 3)) == expect[1:3]
	assert get(None, slice(-30, -3)) == expect[-30:-3]
	assert get(None, -100) == expect
	assert get(None, 0) == expect
	assert get(None, slice(0, None)) == expect
	assert get(None, slice(None, None)) == expect
	assert get(None, 100) == []
	assert get(0, None) == expect[:3]
	assert get(0, -2) == expect[1:3]
	assert get(2, 3) == expect[6:]
	assert get(2, -97) == expect[-97:]
	assert get('roundrobin', 3) == [4, 2] + expect[5:]
	assert get('roundrobin', -40) == expect[-40:]
	assert get('roundrobin', slice(3)) == [0, 3, 1]
	assert get('roundrobin', slice(-10)) == [0, 3, 1, 4, 2] + expect[5:-10]
	assert get('roundrobin', 100) == []
	def assert_fails(sliceno, slice, func=get, columns='a'):
		try:
			func(sliceno, slice, columns)
			raise Exception("Iterating with slice %r should have failed" % (slice,))
		except DatasetError:
			pass
	assert_fails(None, -101)
	assert_fails(None, slice(20, 10))
	assert_fails(None, slice(20, 10, -1))
	assert_fails(None, slice(0, None, -1))
	assert_fails(None, 101)
	assert_fails(0, 4)
	assert_fails(2, -98)
	assert_fails('roundrobin', 101)
	dw = job.datasetwriter({'a': 'int32', 'b': 'int32'}, previous=ds, name='second')
	write = dw.get_split_write()
	write(100, -1)
	write(101, -2)
	write(102, -3)
	ds = dw.finish()
	expect2 = [(100, -1), (101, -2), (102, -3)]
	expect.extend(v[0] for v in expect2)
	assert get(None, None, None) == expect2
	assert get_chain(None, None) == expect
	assert get_chain(0, None) == [0, 1, 2, 100]
	assert get_chain(0, 2) == [2, 100]
	assert get_chain(0, -2) == [2, 100]
	assert get_chain(0, -1, ['a', 'b']) == [(100, -1)] # no error because we slice off 'first'
	assert_fails(0, -2, get_chain, ['a', 'b']) # error because 'first' has no b column.
	assert get_chain(None, slice(90, -2)) == expect[90:-2]
	assert get_chain(None, slice(99, -4)) == []
	assert_fails(None, slice(99, -5), get_chain)
	assert get_chain(None, slice(2, -2, 2)) == expect[2:-2:2]



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify the dataset_hashpart method with various options.
'''

from datetime import date

from accelerator import subjobs
from accelerator.dataset import DatasetWriter, Dataset
from accelerator.dsutil import typed_writer

data = [
	{"a column": "text", "also a column": b"octets", "number": 10, "date": date(1978, 1, 1)},
	{"a column": "a", "also a column": b"0", "number": -900, "date": date(2009, 1, 1)},
	{"a column": "b", "also a column": b"1", "number": -901, "date": date(2009, 1, 2)},
	{"a column": "c", "also a column": b"2", "number": -902, "date": date(2009, 1, 3)},
	{"a column": "d", "also a column": b"3", "number": -903, "date": date(2009, 1, 4)},
	{"a column": "e", "also a column": b"4", "number": -904, "date": date(2009, 1, 5)},
	{"a column": "f", "also a column": b"5", "number": -905, "date": date(2009, 1, 6)},
	{"a column": "g", "also a column": b"6", "number": -906, "date": date(2009, 1, 7)},
	{"a column": "h", "also a column": b"7", "number": -907, "date": date(2009, 1, 8)},
	{"a column": "i", "also a column": b"8", "number": -908, "date": date(2009, 1, 9)},
	{"a column": "j", "also a column": b"9", "number": -909, "date": date(2009, 2, 1)},
	{"a column": "k", "also a column": b"Z", "number": -999, "date": date(2009, 3, 1)},
	{"a column": "l", "also a column": b"Y", "number": -989, "date": date(2009, 4, 1)},
	{"a column": "m", "also a column": b"X", "number": -979, "date": date(2009, 5, 1)},
	{"a column": "n", "also a column": b"W", "number": -969, "date": date(2009, 6, 1)},
	{"a column": "o", "also a column": b"V", "number": -959, "date": date(2009, 7, 1)},
	{"a column": "p", "also a column": b"A", "number": -949, "date": date(2009, 8, 1)},
	{"a column": "q", "also a column": b"B", "number": -939, "date": date(2009, 9, 1)},
	{"a column": "r", "also a column": b"C", "number": -929, "date": None},
	{"a column": "s", "also a column": b"D", "number": None, "date": date(1970, 1, 1)},
	{"a column": "B", "also a column": None, "number": -242, "date": date(1970, 2, 3)},
	{"a column": None, "also a column": b"F", "number": -123, "date": date(1970, 4, 1)},
]
bonus_data = [
	{"a column": "foo", "also a column": b"bar", "number": 42, "date": date(2019, 4, 10)},
]
columns = {
	"a column": ("ascii", True),
	"also a column": ("bytes", True),
	"number": ("int32", True),
	"date": ("date", True),
}

def write(data, **kw):
	dw = DatasetWriter(columns=columns, **kw)
	w = dw.get_split_write_dict()
	for values in data:
		w(values)
	return dw.finish()

def verify(slices, data, source, previous=None, **options):
	jid = subjobs.build(
		"dataset_hashpart",
		datasets=dict(source=source, previous=previous),
		options=options,
	)
	hl = options["hashlabel"]
	h = typed_writer(columns[hl][0]).hash
	ds = Dataset(jid)
	good = {row[hl]: row for row in data}
	names = list(source.columns)
	for slice in range(slices):
		for row in ds.iterate_chain(slice, names):
			row = dict(zip(names, row))
			assert h(row[hl]) % slices == slice, "row %r is incorrectly in slice %d in %s" % (row, slice, ds)
			want = good[row[hl]]
			assert row == want, '%s (rehashed from %s) did not contain the right data for "%s".\nWanted\n%r\ngot\n%r' % (ds, source, hl, want, row)
	return ds

def synthesis(params):
	ds = write(data)
	for colname in data[0]:
		verify(params.slices, data, ds, hashlabel=colname)
	# ok, all the hashing stuff works out, let's test the chaining options.
	bonus_ds = write(bonus_data, name="bonus", previous=ds)
	# no chaining options - full chain
	verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date")
	# just the bonus ds
	verify(params.slices, bonus_data, bonus_ds, hashlabel="date", length=1)
	# built as a chain
	verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date", as_chain=True)
	# normal chaining
	a = verify(params.slices, data, ds, hashlabel="date")
	b = verify(params.slices, data + bonus_data, bonus_ds, hashlabel="date", previous=a)
	assert b.chain() == [a, b], "chain of %s is not [%s, %s] as expected" % (b, a, b)
	# as_chain sparseness
	dw = DatasetWriter(columns=columns, name="empty")
	dw.get_split_write()
	ds = verify(params.slices, [], dw.finish(), hashlabel="date", as_chain=True)
	assert len(ds.chain()) == 1, ds + ": dataset_hashpart on empty dataset with as_chain=True did not produce a single dataset"
	# two populated slices with the same data, should end up in two datasets.
	dw = DatasetWriter(columns=columns, name="0 and 2")
	dw.set_slice(0)
	dw.write_dict(data[0])
	dw.set_slice(1)
	dw.set_slice(2)
	dw.write_dict(data[0])
	for s in range(3, params.slices):
		dw.set_slice(s)
	ds = verify(params.slices, [data[0]], dw.finish(), hashlabel="date", as_chain=True)
	got_slices = len(ds.chain())
	assert got_slices == 2, "%s (built with as_chain=True) has %d datasets in chain, expected 2." % (ds, got_slices,)



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test the dataset_filter_columns method.
'''

from accelerator import subjobs

def synthesis(job):
	dw = job.datasetwriter()
	dw.add('a', 'int32')
	dw.add('b', 'int32')
	dw.add('c', 'int32')
	dw.add('d', 'int32')
	w = dw.get_split_write()
	w(1, 2, 3, 4)
	w(97, 98, 99, 100)
	ds = dw.finish()
	def chk(j, *want):
		ds = j.dataset()
		want = set(want)
		got = set(ds.columns)
		assert got == want, "%s should have had columns %r but had %r" % (ds, want, got,)
		want = list(zip(*[(ord(c) - 96, ord(c)) for c in sorted(want)]))
		got = list(ds.iterate(None))
		assert got == want, "%s should have had %r but had %r" % (ds, want, got,)
	chk(job, 'a', 'b', 'c', 'd')
	j = subjobs.build('dataset_filter_columns', source=ds, keep_columns=['a'])
	chk(j, 'a')
	j = subjobs.build('dataset_filter_columns', source=ds, keep_columns=['b', 'c'])
	chk(j, 'b', 'c')
	j = subjobs.build('dataset_filter_columns', source=ds, discard_columns=['d', 'c', 'b'])
	chk(j, 'a')
	j = subjobs.build('dataset_filter_columns', source=ds, discard_columns=['b'])
	chk(j, 'a', 'c', 'd')
	# Discarding a non-existant column is ok.
	j = subjobs.build('dataset_filter_columns', source=ds, discard_columns=['nah', 'b'])
	chk(j, 'a', 'c', 'd')
	j = subjobs.build('dataset_filter_columns', source=ds, keep_columns=['b', 'b', 'b'])
	chk(j, 'b')
	p1 = j.dataset()
	j = subjobs.build('dataset_filter_columns', source=ds, keep_columns=['a', 'b'], previous=p1)
	chk(j, 'a', 'b')
	p2 = j.dataset()
	assert p2.previous == p1
	j = subjobs.build('dataset_filter_columns', source=p2, keep_columns=['b'])
	chk(j, 'b')
	p3 = j.dataset()
	assert not p3.previous



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
# Modifications copyright (c) 2019 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# parsing of "job specs", including as part of a dataset name.
# handles jobids, paths and method names.

from __future__ import division, print_function

import sys
from os.path import join, exists, realpath, split
from os import readlink, environ
import re

from accelerator.job import WORKDIRS
from accelerator.job import Job
from accelerator.error import NoSuchJobError, NoSuchDatasetError, NoSuchWorkdirError, UrdError
from accelerator.unixhttp import call
from accelerator.compat import url_quote

class JobNotFound(NoSuchJobError):
	pass

class DatasetNotFound(NoSuchDatasetError):
	pass

def _groups(tildes):
	def char_and_count(buf):
		char, count = re.match(r'([~^]+)(\d*)$', ''.join(buf)).groups()
		count = int(count or 1) - 1
		return char[0], len(char) + count
	i = iter(tildes)
	buf = [next(i)]
	for c in i:
		if c in '~^' and buf[-1] != c:
			yield char_and_count(buf)
			buf = [c]
		else:
			buf.append(c)
	yield char_and_count(buf)

# "foo~~^3" -> "foo", [("~", 2), ("^", 3)]
def split_tildes(n):
	m = re.match(r'(.*?)([~^][~^\d]*)$', n)
	if m:
		n, tildes = m.groups()
		lst = list(_groups(tildes))
	else:
		lst = []
	assert n, "empty job id"
	return n, lst

def method2job(cfg, method, count=0, start_from=None):
	def get(count):
		url ='%s/method2job/%s/%s' % (cfg.url, method, count)
		if start_from:
			url += '?start_from=' + url_quote(start_from)
		return call(url)
	found = get(count)
	if 'error' in found:
		raise JobNotFound(found.error)
	return Job(found.id)

# follow jobs.previous (or datasets.previous.job if that is unavailable) count times.
def job_up(job, count):
	err_job = job
	for ix in range(count):
		prev = job.params.jobs.get('previous')
		if not prev:
			prev = job.params.datasets.get('previous')
			if prev:
				prev = prev.job
		if not prev:
			raise JobNotFound('Tried to go %d up from %s, but only %d previous jobs available' % (count, err_job, ix,))
		job = prev
	return job

def urd_call_w_tildes(cfg, path, tildes):
	res = call(cfg.urd + '/' + path, server_name='urd', retries=0, quiet=True)
	if tildes:
		up = sum(count for char, count in tildes if char == '^')
		down = sum(count for char, count in tildes if char == '~')
		tildes = down - up
		if tildes:
			key = res.user + '/' + res.build
			timestamps = call(cfg.urd + '/' + key + '/since/0', server_name='urd', retries=0, quiet=True)
			pos = timestamps.index(res.timestamp) + tildes
			if pos < 0 or pos >= len(timestamps):
				return None
			res = call(cfg.urd + '/' + key + '/' + timestamps[pos], server_name='urd', retries=0, quiet=True)
	return res

def name2job(cfg, n):
	n, tildes = split_tildes(n)
	job = _name2job(cfg, n)
	for char, count in tildes:
		if char == '~':
			job = method2job(cfg, job.method, count, start_from=job)
		else:
			job = job_up(job, count)
	return job

def _name2job(cfg, n):
	if n.startswith(':'):
		# resolve through urd
		assert cfg.urd, 'No urd configured'
		a = n[1:].rsplit(':', 1)
		if len(a) == 1:
			raise JobNotFound('looks like a partial :urdlist:[entry] spec')
		entry = a[1] or '-1'
		try:
			entry = int(entry, 10)
		except ValueError:
			pass
		path, tildes = split_tildes(a[0])
		path = path.split('/')
		if len(path) < 3:
			path.insert(0, environ.get('USER', 'NO-USER'))
		if len(path) < 3:
			path.append('latest')
		path = '/'.join(map(url_quote, path))
		try:
			urdres = urd_call_w_tildes(cfg, path, tildes)
		except UrdError as e:
			print(e, file=sys.stderr)
			urdres = None
		if not urdres:
			raise JobNotFound('urd list %r not found' % (a[0],))
		from accelerator.build import JobList
		joblist = JobList(Job(e[1], e[0]) for e in urdres.joblist)
		res = joblist.get(entry)
		if not res:
			raise JobNotFound('%r not found in %s' % (entry, path,))
		return res
	if re.match(r'[^/]+-\d+$', n):
		# Looks like a jobid
		return Job(n)
	m = re.match(r'([^/]+)-LATEST$', n)
	if m:
		# Looks like workdir-LATEST
		wd = m.group(1)
		if wd not in WORKDIRS:
			raise NoSuchWorkdirError('Not a valid workdir: "%s"' % (wd,))
		path = join(WORKDIRS[wd], n)
		try:
			n = readlink(path)
		except OSError as e:
			raise JobNotFound('Failed to read %s: %s' % (path, e,))
		return Job(n)
	if '/' not in n:
		# Must be a method then
		return method2job(cfg, n)
	if exists(join(n, 'setup.json')):
		# Looks like the path to a jobdir
		path, jid = split(realpath(n))
		job = Job(jid)
		if WORKDIRS.get(job.workdir, path) != path:
			print("### Overriding workdir %s to %s" % (job.workdir, path,))
		WORKDIRS[job.workdir] = path
		return job
	raise JobNotFound("Don't know what to do with %r." % (n,))

def name2ds(cfg, n):
	job = name = tildes = None
	try:
		job = name2job(cfg, n)
	except JobNotFound:
		if '/' not in n:
			raise
	if not job:
		n, name = n.rsplit('/', 1)
		job = name2job(cfg, n)
		name, tildes = split_tildes(name)
	ds = job.dataset(name)
	if tildes:
		def follow(key, motion):
			# follow ds.key count times
			res = ds
			for done in range(count):
				if not getattr(res, key):
					raise DatasetNotFound('Tried to go %d %s from %s, but only %d available (stopped on %s)' % (count, motion, ds, done, res,))
				res = getattr(res, key)
			return res
		for char, count in tildes:
			if char == '~':
				ds = follow('previous', 'back')
			else:
				ds = follow('parent', 'up')
	slices = ds.job.params.slices
	from accelerator import g
	if hasattr(g, 'slices'):
		assert g.slices == slices, "Dataset %s needs %d slices, by we are already using %d slices" % (ds, slices, g.slices)
	else:
		g.slices = slices
	return ds



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from collections import OrderedDict
from json import dumps
from datetime import datetime, date, time, timedelta

from accelerator.compat import iteritems, unicode, long, PY3, PY2, uni

from accelerator.extras import DotDict, json_load, json_save, json_encode
from accelerator.job import Job

def generate(caption, method, options=None, datasets=None, jobs=None, package=None, description=None, why_build=False):
	data = DotDict()
	data.caption = caption
	data.method  = method
	if package:
		data.package = package
	data.versions = DotDict()
	if description:
		data.versions.python_path = description['interpreter']
		data.versions.python = description['interpreter_version']
		if description['accelerator_version']:
			data.versions.accelerator = description['accelerator_version']
	if why_build:
		data.why_build = why_build
	data.options = options or DotDict()
	data.datasets = datasets or DotDict()
	data.jobs = jobs or DotDict()
	return data

def load_setup(jobid):
	"""Loads but does not type setup.json from jobid.
	You probably want to use extras.job_params instead.
	"""
	d = json_load('setup.json', jobid)
	version = d.version
	if version == 1:
		d.jobs = d.pop('jobids')
		version = 2
	if version == 2:
		if 'exectime' not in d and 'profile' in d:
			d.exectime = d.pop('profile')
		d.versions = DotDict()
		python_path = d.pop('python', None)
		if python_path:
			d.versions.python_path = python_path
		version = 3
	if version != 3:
		raise Exception("Don't know how to load setup.json version %d (in %s)" % (d.version, jobid,))
	return d

def update_setup(jobid, **kw):
	data = load_setup(jobid)
	data.update(kw)
	save_setup(jobid, data)
	return data

# It's almost worth making your own json encoder. Almost.
def _sorted_set(s):
	# like sorted(s), except None is ok.
	if None in s:
		s = set(s)
		s.remove(None)
		res = sorted(s)
		res.append(None)
		return res
	else:
		return sorted(s)

def encode_setup(data, sort_keys=True, as_str=False):
	def copy(src):
		if isinstance(src, dict):
			dst = OrderedDict()
			for k in sorted(src):
				dst[k] = copy(src[k])
			return dst
		elif isinstance(src, (list, tuple,)):
			return [copy(v) for v in src]
		elif isinstance(src, set):
			return [copy(v) for v in _sorted_set(src)]
		elif isinstance(src, datetime):
			return [src.year, src.month, src.day, src.hour, src.minute, src.second, src.microsecond]
		elif isinstance(src, date):
			return [src.year, src.month, src.day]
		elif isinstance(src, time):
			return [1970, 1, 1, src.hour, src.minute, src.second, src.microsecond]
		elif isinstance(src, timedelta):
			return src.total_seconds()
		elif PY2 and isinstance(src, bytes):
			return uni(src)
		else:
			assert isinstance(src, (str, unicode, int, float, long, bool)) or src is None, type(src)
			return src
	res = _encode_with_compact(
		copy(data),
		compact_keys=('starttime', 'endtime', 'exectime', '_typing',),
		special_keys=('options', 'datasets', 'jobs',),
	)
	if PY3 and not as_str:
		res = res.encode('ascii')
	return res

def _round_floats(d, ndigits):
	res = OrderedDict()
	for k, v in iteritems(d):
		if isinstance(v, float):
			v = round(v, ndigits)
		if isinstance(v, dict):
			v = _round_floats(v, ndigits)
		if isinstance(v, list):
			v = [round(e, ndigits) if isinstance(e, float) else e for e in v]
		res[k] = v
	return res

def _encode_with_compact(data, compact_keys, extra_indent=0, separator='\n', special_keys=()):
	compact = []
	special = []
	for k in compact_keys:
		if k in data:
			if k == 'exectime':
				d = _round_floats(data[k], 3)
				fmted = _encode_with_compact(d, ('analysis', 'per_slice',), 1, '')
			else:
				fmted = dumps(data[k])
			compact.append('    "%s": %s,' % (k, fmted,))
			del data[k]
	for k in special_keys:
		if k in data:
			fmted = dumps(data[k], indent=4, sort_keys=True)
			special.append('    "%s": %s' % (k, fmted.replace('\n', '\n    '),))
			del data[k]
	res = json_encode(data, as_str=True)
	if compact:
		res = '{\n%s%s%s' % ('\n'.join(compact), separator, res[1:],)
	if special:
		res = '%s,\n\n%s\n}' % (res[:-2], ',\n'.join(special),)
	res = res.replace('\n', ('\n' + ' ' * extra_indent * 4))
	return res

def save_setup(jobid, data):
	data = dict(data)
	data['version'] = 3
	filename = Job(jobid).filename('setup.json')
	json_save(data, filename, _encoder=encode_setup)



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test DatasetWriter, exercising the different ways to create,
pass and populate the dataset.
'''

from datetime import date

from accelerator.dataset import DatasetWriter
from . import test_data

depend_extra=(test_data,)

def prepare(job, slices):
	assert slices >= test_data.value_cnt
	dw_default = DatasetWriter()
	dw_default.add("a", "number")
	dw_default.add("b", "ascii")
	DatasetWriter(name="named", columns={"c": "bool", "d": "date"})
	dw_passed = job.datasetwriter(name="passed", columns=test_data.columns)
	return dw_passed, 42

def analysis(sliceno, prepare_res, job):
	dw_default = job.datasetwriter()
	dw_named = DatasetWriter(name="named")
	dw_passed, num = prepare_res
	dw_default.write(a=sliceno, b="a")
	dw_default.write_list([num, str(sliceno)])
	dw_named.write(True, date(1536, 12, min(sliceno + 1, 31)))
	dw_named.write_dict({"c": False, "d": date(2236, 5, min(sliceno + 1, 31))})
	# slice 0 is written in synthesis
	if 0 < sliceno < test_data.value_cnt:
		dw_passed.write_dict({k: v[sliceno] for k, v in test_data.data.items()})

def synthesis(prepare_res, slices, job):
	dw_passed, _ = prepare_res
	# Using set_slice on a dataset that was written in analysis is not
	# actually supported, but since it currently works (as long as that
	# particular slice wasn't written in analysis) let's test it.
	dw_passed.set_slice(0)
	dw_passed.write(**{k: v[0] for k, v in test_data.data.items()})
	dw_synthesis_split = DatasetWriter(name="synthesis_split", hashlabel="a")
	dw_synthesis_split.add("a", "int32")
	dw_synthesis_split.add("b", "unicode")
	dw_synthesis_split.get_split_write()(1, "a")
	dw_synthesis_split.get_split_write_list()([2, "b"])
	dw_synthesis_split.get_split_write_dict()({"a": 3, "b": "c"})
	dw_synthesis_manual = job.datasetwriter(name="synthesis_manual", columns={"sliceno": "int32"})
	dw_nonetest = job.datasetwriter(name="nonetest", columns=test_data.columns)
	for sliceno in range(slices):
		dw_synthesis_manual.set_slice(sliceno)
		dw_synthesis_manual.write(sliceno)
		dw_nonetest.set_slice(sliceno)
		dw_nonetest.write(**{k: v[0] if k in test_data.not_none_capable else None for k, v in test_data.data.items()})



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that NaN does not end up in min/max unless it's the only value.
'''

from math import isnan

nan = float('nan')
inf = float('inf')

def prepare(job):
	def create(name, none_support=False):
		dw = job.datasetwriter(name=name)
		dw.add('float32', 'float32', none_support=none_support)
		dw.add('float64', 'float64', none_support=none_support)
		dw.add('number', 'number', none_support=none_support)
		return dw
	return create('a', True), create('b'), create('c')

def analysis(sliceno, prepare_res):
	a, b, c = prepare_res
	if sliceno < 2:
		for ds in prepare_res:
			ds.write(nan, nan, nan)
		if sliceno == 1:
			a.write(None, None, None)
	if sliceno == 2:
		b.write(nan, 2, nan)
		c.write(nan, nan, nan)
		c.write(inf, 1, 1)
		c.write(0, 1, 2)
		c.write(nan, nan, nan)

def synthesis(prepare_res):
	a, b, c = prepare_res
	def eq(a, b):
		if isinstance(a, float) and isinstance(b, float) and isnan(a) and isnan(b):
			return True
		return a == b
	def check(dw, want_min, want_max):
		ds = dw.finish()
		for colname, want_min, want_max in zip(['float32', 'float64', 'number'], want_min, want_max):
			col = ds.columns[colname]
			assert eq(col.min, want_min), "%s.%s should have had min value %s, but had %s" % (ds, colname, want_min, col.min)
			assert eq(col.max, want_max), "%s.%s should have had max value %s, but had %s" % (ds, colname, want_max, col.max)
	check(a, [nan, nan, nan], [nan, nan, nan])
	check(b, [nan, 2, nan], [nan, 2, nan])
	check(c, [0, 1, 1], [inf, 1, 2])



############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset_sort with various options on a dataset with all types.
'''

from itertools import chain
from math import isnan
from functools import cmp_to_key

from accelerator.dataset import Dataset
from accelerator import subjobs
from . import test_data

# ideally the subjobs should be part of staleness detection, but currently
# they are not, so we put gendata here explicitly.
depend_extra = (test_data, "a_test_sorting_gendata.py",)

def noneninf(v):
	if v is None:
		return float('-inf')
	return v

def unnan(l):
	return [
		tuple('NaN' if isinstance(v, float) and isnan(v) else v for v in t)
	for t in l]

def check_one(slices, key, source, reverse=False):
	jid = subjobs.build(
		"dataset_sort",
		options=dict(
			sort_columns=key,
			sort_order="descending" if reverse else "ascending",
		),
		datasets=dict(source=source),
	)
	ds = Dataset(jid)
	key_off = sorted(test_data.data).index(key)
	# This provides better separation than the replacement values
	# used in the actual sort method (but this is slow).
	if 'date' in key or 'time' in key:
		nonepos = 1
	else:
		nonepos = -1
	def cmp(a, b):
		a = a[key_off]
		b = b[key_off]
		if a is None:
			if b is None:
				return 0
			return nonepos
		if b is None:
			return -nonepos
		if isinstance(a, float):
			if isnan(a):
				if isnan(b):
					return 0
				return 1
			if isnan(b):
				return -1
		if a < b:
			return -1
		return a > b
	keycmp = cmp_to_key(cmp)
	for sliceno in range(slices):
		good = sorted(test_data.sort_data_for_slice(sliceno), key=keycmp, reverse=reverse)
		check = list(ds.iterate(sliceno))
		assert unnan(check) == unnan(good), "Slice %d sorted on %s bad (%s)" % (sliceno, key, jid,)

def synthesis(params):
	source = Dataset(subjobs.build("test_sorting_gendata"))
	# Test that all datatypes work for sorting
	for key in test_data.data:
		if not key.startswith('complex'):
			check_one(params.slices, key, source)
	# Check reverse sorting
	check_one(params.slices, "int32", source, reverse=True)
	# Check that sorting across slices and by two columns works
	jid = subjobs.build(
		"dataset_sort",
		options=dict(
			sort_columns=["int64", "int32"],
			sort_order="descending",
			sort_across_slices=True,
		),
		datasets=dict(source=source),
	)
	int64_off = sorted(test_data.data).index("int64")
	int32_off = sorted(test_data.data).index("int32")
	all_data = chain.from_iterable(test_data.sort_data_for_slice(sliceno) for sliceno in range(params.slices))
	good = sorted(all_data, key=lambda t: (noneninf(t[int64_off]), noneninf(t[int32_off]),), reverse=True)
	ds = Dataset(jid)
	check = list(ds.iterate(None))
	assert unnan(check) == unnan(good), "Sorting across slices on [int64, int32] bad (%s)" % (jid,)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import

description = r'''
Rewrite a dataset (or chain to previous) with new hashlabel.
'''

from shutil import copyfileobj
from os.path import exists

from accelerator import OptionString

options = {
	'hashlabel'                 : OptionString,
	'caption'                   : '"%(caption)s" hashed on %(hashlabel)s',
	'length'                    : -1, # Go back at most this many datasets. You almost always want -1 (which goes until previous.source)
	'as_chain'                  : False, # one dataset per slice (avoids rewriting at the end)
}

datasets = ('source', 'previous',)

def prepare(job, slices):
	d = datasets.source
	caption = options.caption % dict(caption=d.caption, hashlabel=options.hashlabel)
	chain = d.chain(stop_ds={datasets.previous: 'source'}, length=options.length)
	if len(chain) == 1:
		filename = d.filename
	else:
		filename = None
	dws = []
	previous = datasets.previous
	if options.as_chain:
		# The last slice that actually has data in it becomes 'default'.
		# Or slice 0 if the whole source is empty (we must produce a ds).
		default_sliceno = 0
		for sliceno in range(slices - 1, -1, -1):
			if chain.lines(sliceno):
				default_sliceno = sliceno
				break
	else:
		default_sliceno = None
	for sliceno in range(slices):
		if sliceno == default_sliceno or chain.lines(sliceno):
			if sliceno == default_sliceno:
				name = "default"
			else:
				name = str(sliceno)
			dw = job.datasetwriter(
				caption="%s (slice %d)" % (caption, sliceno),
				hashlabel=options.hashlabel,
				filename=filename,
				previous=previous,
				name=name,
				for_single_slice=sliceno,
				copy_mode=True,
			)
			previous = dw
		else:
			dw = None
		dws.append(dw)
	names = []
	cols = {}
	for n, c in d.columns.items():
		# names has to be in the same order as the add calls
		# so the iterator returns the same order the writer expects.
		names.append(n)
		cols[n] = (c.type, chain.none_support(n))
		for dw in dws:
			if dw:
				dw.add(n, c.type, none_support=cols[n][1])
	return dws, names, caption, filename, cols

def analysis(sliceno, prepare_res):
	dws, names = prepare_res[:2]
	if not dws[sliceno]:
		return
	it = datasets.source.iterate_chain(
		sliceno,
		names,
		stop_ds={datasets.previous: 'source'},
		length=options.length,
		copy_mode=True,
	)
	write = dws[sliceno].get_split_write_list()
	for values in it:
		write(values)

def synthesis(prepare_res, job, slices):
	if not options.as_chain:
		# If we don't want a chain we abuse our knowledge of dataset internals
		# to avoid recompressing. Don't do this stuff yourself.
		dws, names, caption, filename, cols = prepare_res
		merged_dw = job.datasetwriter(
			caption=caption,
			hashlabel=options.hashlabel,
			filename=filename,
			previous=datasets.previous,
			meta_only=True,
			columns=cols,
		)
		dws = list(filter(None, dws))
		for sliceno in range(slices):
			merged_dw.set_lines(sliceno, sum(dw._lens[sliceno] for dw in dws))
			for dwno, dw in enumerate(dws):
				merged_dw.set_minmax((sliceno, dwno), dw._minmax[sliceno])
			for n in names:
				fn = merged_dw.column_filename(n, sliceno=sliceno)
				with open(fn, "wb") as out_fh:
					for dw in dws:
						fn = dw.column_filename(n, sliceno=sliceno)
						if exists(fn):
							with open(fn, "rb") as in_fh:
								copyfileobj(in_fh, out_fh)
		for dw in dws:
			dw.discard()



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from accelerator import subjobs
from accelerator.error import DatasetUsageError

def synthesis(job):
	dw = job.datasetwriter()
	dw.add('', 'number')
	dw.add('word', 'ascii')
	w = dw.get_split_write()
	w(0, 'foo')
	w(1, 'bar')
	ds = dw.finish()
	assert set(ds.columns) == {'', 'word'}
	assert list(ds.iterate(None, '')) == [0, 1]
	assert list(ds.iterate(None)) == [(0, 'foo'), (1, 'bar')]
	jid = subjobs.build('csvexport', source=ds, filename='out.csv')
	jid = subjobs.build('csvimport', filename=jid.filename('out.csv'))
	jid = subjobs.build('dataset_type', source=jid, column2type={'': 'number', 'word': 'ascii'})
	assert list(jid.dataset().iterate(None)) == list(ds.iterate(None))

	# verify that names really get resolved/used in added order, for all orders
	# and sneak in a hashlabel='' check as well.
	# this should really be in test_dataset_column_names, but I wanted to
	# include the hashlabel='' check too, so here it is.
	def chk_order(names, want):
		dw = job.datasetwriter(name='order' + ','.join(names), hashlabel='')
		for name in names:
			dw.add(name, 'ascii')
		w = dw.get_split_write()
		# write in the specified order
		w(*names)
		# write with the expected (mangled) name
		w(**{v: k for k, v in want.items()})
		ds = dw.finish()
		for col in want:
			# '' hashes to 0, so if the hashlabel worked both are in slice 0.
			assert list(ds.iterate(0, col)) == [col, col], '%r bad in %s' % (col, ds,)
		return ds
	chk_order(['@', '_', ''], {'@': '_', '_': '__', '': '___'})
	chk_order(['@', '', '_'], {'@': '_', '': '__', '_': '___'})
	chk_order(['', '@', '_'], {'': '_', '@': '__', '_': '___'})
	chk_order(['_', '@', ''], {'_': '_', '@': '__', '': '___'})
	ds = chk_order(['', '_', '@'], {'': '_', '_': '__', '@': '___'})

	# check that hashlabel '' is correctly handled in the mismatch check
	def chk_mismatch(ds, good, bad):
		list(ds.iterate(0, '_', hashlabel=good))
		try:
			list(ds.iterate(0, '_', hashlabel=bad))
		except DatasetUsageError:
			pass
		else:
			raise Exception('%s accepted hashlabel %r' % (ds, bad,))
	chk_mismatch(ds, '', '@')
	dw = job.datasetwriter(name='hl_', hashlabel='_')
	dw.add('_', 'int32')
	dw.add('', 'int32')
	w = dw.get_split_write()
	for ix in range(64):
		w(ix, 0)
	ds = dw.finish()
	chk_mismatch(ds, '_', '')

	# check that re-hashing on '' works
	# 0 hashes to 0, so all should be in slice 0
	got = set(ds.iterate(0, '_', hashlabel='', rehash=True))
	want = set(range(64))
	assert got == want, 'Wanted {0..63}, got %r from %s' % (got, ds,)



############################################################################
#                                                                          #
# Copyright (c) 2020-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test some strange choices for separators in csvexport.
'''

import gzip

from accelerator import subjobs

def synthesis(job):
	dw = job.datasetwriter()
	dw.add('a', 'ascii')
	dw.add('b', 'unicode')
	w = dw.get_split_write()
	w('A', 'B')
	w('\0', '\xe4')
	ds = dw.finish()
	def verify(data, filename):
		want = []
		for line in [['a', 'b']] + data:
			want.append(separator.join(quote + item + quote for item in line))
			want.append(line_separator)
		want = ''.join(want).encode('utf-8')
		if ext == '.gz':
			open_func = gzip.open
		else:
			open_func = open
		with open_func(j.filename(filename), 'rb') as fh:
			got = fh.read()
		assert want == got, "Expected %s/%s to contain %r, but contained %r" % (j, filename, want, got,)
	for separator in ('', '\0', 'wheeee'):
		for line_separator in ('', '\0', 'woooooo'):
			for quote in ('', 'qqq'):
				for ext in ('.csv', '.gz'):
					for sliced, filename in ((False, 'out' + ext), (True, 'out.%d' + ext)):
						j = subjobs.build(
							'csvexport',
							filename=filename,
							separator=separator,
							line_separator=line_separator,
							quote_fields=quote,
							lazy_quotes=False,
							sliced=sliced,
							source=ds,
						)
						if sliced:
							for sliceno, data in ((0, ['A', 'B']), (1, ['\0', '\xe4'])):
								verify([data], filename % (sliceno,))
						else:
							verify([['A', 'B'], ['\0', '\xe4']], filename)
